# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2009 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Ronaldo Maia <romaia@async.com.br>
##
"""
Bematech MP20 driver

The MP20 is compatible with the MP25 command set (actually its the other way
around ;) until a certain command (85, I think). Commands above that are just
not executed.

There are some differences on the Registers numbering as well.

Also, some commands have different parameter sizes. These are:

CMD             MP 20                MP 25
00              29                  28+30+80 (abertura de cupom)
14              -                   28+30+80 (cancelamento de cupom)
32                                           (inicia fechamento cupom)
73              Algumas diferencas no funcionamento. Ver manual.
"""

from decimal import Decimal
import logging

from stoqdrivers.printers.bematech.MP25 import (MP25, MP25Status, CMD_STATUS,
                                                CMD_COUPON_OPEN)

log = logging.getLogger('stoqdrivers.bematech.MP20')

CMD_ADD_ITEM_SIMPLE = 9


class MP20Registers(object):
    TOTAL = 3
    TOTAL_CANCELATIONS = 4
    TOTAL_DISCOUNT = 5
    COO = 6
    GNF = 7
    NUMBER_REDUCTIONS_Z = 9
    CRO = 10
    LAST_ITEM_ID = 12
    NUMBER_TILL = 14
    FISCAL_FLAGS = 17
    EMISSION_DATE = 23
    TOTALIZERS = 29
    PAYMENT_METHODS = 32
    SERIAL = 0
    FIRMWARE = 1

    # (size, bcd)
    formats = {
        TOTAL: ('9s', True),
        TOTAL_CANCELATIONS: ('7s', True),
        TOTAL_DISCOUNT: ('7s', True),
        COO: ('3s', True),
        GNF: ('3s', True),
        NUMBER_REDUCTIONS_Z: ('2s', True),
        CRO: ('2s', True),
        LAST_ITEM_ID: ('2s', True),
        NUMBER_TILL: ('2s', True),
        FISCAL_FLAGS: ('1s', False),
        EMISSION_DATE: ('6s', False),
        TOTALIZERS: ('2s', False),
        #  1 + (52 * 16) + (52 * 10) + (52 * 10) + (52 * 1)
        #  1 + 832 + 520 + 520 + 52: 1925
        PAYMENT_METHODS: ('b832s520s520s52s', False),
        SERIAL: ('15s', False),
        FIRMWARE: ('3s', True),
    }


class MP20Status(MP25Status):
    def __init__(self, reply):
        self.st1, self.st2 = reply[-2:]
        self.st3 = 0


class MP20(MP25):
    model_name = "Bematech MP20 TH FI"
    CMD_PROTO = 0x1b

    registers = MP20Registers
    supports_duplicate_receipt = False
    reply_format = '<b%sbb'
    status_size = 2

    #
    #   MP25 implementation
    #

    def coupon_open(self):
        """ This needs to be called before anything else. """
        self._send_command(CMD_COUPON_OPEN,
                           "%-29s" % (self._customer_document))

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=None,
                        discount=Decimal("0.0"), markup=Decimal("0.0"),
                        unit_desc=""):

        # We are using a simpler command for adding items with the MP20
        # because its not working with the MP25 command (ESC 63). This
        # simpler command does not support markup and unit
        data = (
            "%-13s"  # code
            "%29s"  # description
            "%02s"     # taxcode
            "%07d"     # quantity
            "%08d"     # value
            "%08d"    # discount
        ) % (code, description, taxcode, quantity * Decimal("1e3"),
             price * Decimal("1e2"), discount * Decimal("1e2"))

        self._send_command(CMD_ADD_ITEM_SIMPLE, data)
        return self._get_last_item_id()

    def get_status(self, val=None):
        if val is None:
            val = self._send_command(CMD_STATUS, raw=True)

        return MP20Status(val)

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        #XXX MP20 does not support this
        self.coupon_cancel()

    def get_ccf(self):
        # MP20 does not support this. We should just return the coo
        # http://www.forumweb.com.br/foruns/lofiversion/index.php/t64417.html
        return self.get_coo()

    def status_reply_complete(self, reply):
        log.debug('status_reply_complete "%s" (size=%s)' % (reply, len(reply)))
        return len(reply) == 18



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

from zope.interface import implements

from stoqdrivers.interfaces import INonFiscalPrinter
from stoqdrivers.serialbase import SerialBase

NORMAL_FONT = '\x1b\x21\x00\x00'

CONDENSED_MODE = '\x1b\x0F\x00'
CENTRALIZE = '\x1b\x6A\x01'
DESCENTRALIZE = '\x1b\x6A\x00'
SET_BOLD = '\x1b\x45\x00'
UNSET_BOLD = '\x1b\x46\x00'


class DR700(SerialBase):
    implements(INonFiscalPrinter)

    supported = True
    model_name = "Daruma DR 700"

    def __init__(self, port, consts=None):
        self._is_bold = False
        self._is_centralized = False
        SerialBase.__init__(self, port)
        self.write(NORMAL_FONT)
        self.write(CONDENSED_MODE)

    def centralize(self):
        if self._is_centralized:
            return
        self.write(CENTRALIZE)
        self._is_centralized = True

    def descentralize(self):
        if not self._is_centralized:
            return
        self.write(DESCENTRALIZE)
        self._is_centralized = False

    def set_bold(self):
        if self._is_bold:
            return
        self.write(SET_BOLD)
        self._is_bold = True

    def unset_bold(self):
        if not self._is_bold:
            return
        self.write(UNSET_BOLD)
        self._is_bold = False

    def print_line(self, data):
        self.write(data + '\n')

    def print_inline(self, data):
        self.write(data)

    def print_barcode(self, code):
        code_128 = chr(5)
        width = chr(2)
        height = chr(80)
        barcode_label = chr(0)
        self.write('\x1b\x62%s%s%s%s%s\x00' % (code_128, width, height,
                                               barcode_label, code))
        self.write('\x0A')

    def print_qrcode(self, code):
        bytes_len = len(code)
        min_size = chr(bytes_len >> 8)
        max_size = chr((bytes_len & 255) + 2)
        width = chr(3)
        # Correction level: auto
        correction = chr(0)
        self.write('\x1b\x81%s%s%s%s%s' % (max_size, min_size, width, correction, code))
        self.write('\x0A')



# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2015 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): Stoq Team <stoq-devel@async.com.br>
##

import qrcode
from zope.interface import implements

from stoqdrivers.interfaces import INonFiscalPrinter
from stoqdrivers.serialbase import SerialBase

LINE_FEED = '\x1bA\x00'
CENTRALIZE = '\x1ba\x01'
DESCENTRALIZE = '\x1ba\x00'
CONDENSED_MODE = '\x1bSI'
NORMAL_MODE = '\x1bH'
SET_BOLD = '\x1bE'
UNSET_BOLD = '\x1bF'
BARCODE_128 = '\x1Dkn'

_GRAPHICS_8BITS = 8
_GRAPHICS_24BITS = 24
_GRAPHICS_MAX_COLS = {
    _GRAPHICS_8BITS: 576,
    _GRAPHICS_24BITS: 1728,
}
_GRAPHICS_CMD = {
    _GRAPHICS_8BITS: '\x1b\x4b%s%s%s',
    _GRAPHICS_24BITS: '\x1b\x2a\x21%s%s%s',
}


def _bits2byte(bits):
    return sum(2 ** i if bit else 0 for i, bit in enumerate(reversed(bits)))


def _matrix2graphics(graphics_api, matrix, multiplier=1, centralized=True):
    if not graphics_api in [_GRAPHICS_8BITS, _GRAPHICS_24BITS]:
        raise ValueError("Graphics api %s not supported" % (graphics_api, ))

    sub_len = graphics_api / multiplier

    for i in xrange(0, len(matrix), sub_len):
        bytes_ = []
        sub = matrix[i:i + sub_len]
        if len(sub) < sub_len:
            sub.extend([[False] * len(matrix)] * (sub_len - len(sub)))

        for j in xrange(len(matrix)):
            bits = []
            for bit in sub:
                bits.extend([bit[j]] * multiplier)

            if graphics_api == _GRAPHICS_8BITS:
                # The 3 is to compensate for the fact that each pixel is
                # 3x larger vertically than horizontally
                bytes_.extend([_bits2byte(bits)] * 3 * multiplier)
            elif graphics_api == _GRAPHICS_24BITS:
                splitted_bytes = []
                for k in xrange(0, 24, 8):
                    splitted_bytes.append(_bits2byte(bits[k: k + 8]))
                bytes_.extend(splitted_bytes * multiplier)
            else:
                raise AssertionError

        if centralized:
            diff = _GRAPHICS_MAX_COLS[graphics_api] - len(bytes_)
            if diff:
                bytes_ = ([0] * (diff / 2)) + bytes_

        divide_len_by = graphics_api / 8
        yield ''.join(chr(b) for b in bytes_), len(bytes_) / divide_len_by


class MP2100TH(SerialBase):
    implements(INonFiscalPrinter)

    supported = True
    model_name = "Bematech MP2100 TH"

    def __init__(self, port, consts=None):
        self._is_bold = False
        self._is_centralized = False
        SerialBase.__init__(self, port)
        self.write(CONDENSED_MODE)

    #
    #  INonFiscalPrinter
    #

    def centralize(self):
        if self._is_centralized:
            return
        self.write(CENTRALIZE)
        self._is_centralized = True

    def descentralize(self):
        if not self._is_centralized:
            return
        self.write(DESCENTRALIZE)
        self._is_centralized = False

    def set_bold(self):
        if self._is_bold:
            return
        self.write(SET_BOLD)
        self._is_bold = True

    def unset_bold(self):
        if not self._is_bold:
            return
        self.write(UNSET_BOLD)
        self._is_bold = False

    def print_line(self, data):
        self.write(data + '\n')

    def print_inline(self, data):
        self.write(data)

    def print_barcode(self, code):
        # Change the height
        self.write('\x1d\x68%s' % chr(120))
        # Normal width
        self.write('\x1d\x77\x02')
        # No HRI (human readable information)
        self.write('\x1d\x48\x00')

        cmd = '\x1d\x6b\x49%s%s' % (chr(len(code)), code)
        self.write(cmd)

    def print_qrcode(self, code):
        qr = qrcode.QRCode(version=1, border=4)
        qr.add_data(code)
        self.write('\x00')
        self._print_matrix(_GRAPHICS_8BITS, qr.get_matrix())

    #
    #  Private
    #

    def _print_matrix(self, graphics_api, matrix, multiplier=1):
        for line, line_len in _matrix2graphics(graphics_api, matrix, multiplier):
            assert line_len <= _GRAPHICS_MAX_COLS[graphics_api]
            n2 = 0
            n1 = line_len
            # line_len = n1 + n2 * 256
            while n1 >= 256:
                n2 += 1
                n1 -= 256

            self.write(_GRAPHICS_CMD[graphics_api] % (chr(n1), chr(n2), line))
            self.write(LINE_FEED)



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano  <henrique@async.com.br>
##
"""
Barcode reader interface implementation.
"""

from zope.interface import providedBy

from stoqdrivers.interfaces import IBarcodeReader
from stoqdrivers.base import BaseDevice
from stoqdrivers.enum import DeviceType


class BarcodeReader(BaseDevice):
    device_dirname = "readers.barcode"
    device_type = DeviceType.BARCODE_READER

    def check_interfaces(self):
        if not IBarcodeReader in providedBy(self._driver):
            raise TypeError("The driver `%r' doesn't implements a valid "
                            "interface" % self._driver)

    def get_code(self):
        return self._driver.get_code()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##
"""
Dataregis EP 275 drivers
"""

import datetime
from decimal import Decimal
import time

from zope.interface import implements

from stoqdrivers.serialbase import SerialBase
from stoqdrivers.interfaces import (IChequePrinter,
                                    ICouponPrinter)
from stoqdrivers.exceptions import (DriverError, PendingReduceZ, PendingReadX,
                                    PrinterError, CommError, CommandError,
                                    CommandParametersError, ReduceZError,
                                    HardwareFailure, OutofPaperError,
                                    CouponNotOpenError, CancelItemError,
                                    CouponOpenError)
from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.printers.cheque import (BaseChequePrinter,
                                         BankConfiguration)
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.translation import stoqdrivers_gettext

EOT = 0x04
BS = 0x08
ACK = 0x06
CR = 0x0D
SUB = 0x1A

_ = stoqdrivers_gettext

#
# Helper functions
#


def format_value(value, max_len):
    value = "%-*.02f" % (max_len, value)
    if len(value) > max_len:
        raise ValueError("The value is too big")
    return value


class EP375Constants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: '00',
        UnitType.METERS: '04',
        UnitType.LITERS: '03',
        UnitType.EMPTY: '02',
        PaymentMethodType.MONEY: '00',
        PaymentMethodType.CHECK: '01',
    }

    _tax_constants = [
        (TaxType.SUBSTITUTION, '02', None),
        (TaxType.EXEMPTION, '03', None),
        (TaxType.NONE, '04', None),
    ]


#
# Class implementation to printer status management
#

class EP375Status:
    opened_drawer = None
    has_cmc7 = None
    technical_mode = None
    internal_state = None
    is_ready = None
    statuses = None

    NEEDS_REDUCE_Z = 0x4F
    NEEDS_READ_X = 0x41
    HAS_BEEN_TOTALIZED = 0x46
    PRINTER_IS_OK = 0x4B
    HAS_OPENED_REPORT = 0x52
    HAS_NO_FISCAL_SALE = 0x49
    HAS_FISCAL_SALE = 0x56

    errors_dict = {
        0x41: (PrinterError, _("Fiscal memory has changed")),
        0x61: (PrinterError, _("No manufacture number")),
        0x42: (CommError, _("Print buffer is full")),
        0x62: (CancelItemError, _("No item(s) to cancel found")),
        0x43: (CommandError, _("The requested command doesn't exist")),
        0x63: (DriverError, "Cancellation above the limit"),
        0x44: (DriverError, "Discount more than total value"),
        0x64: (DriverError, "Invalid date"),
        0x45: (HardwareFailure, _("Fiscal EPROM disconnected")),
        0x65: (PrinterError, _("Incorrect version of the basic software")),
        0x46: (PrinterError, _("Error on the fiscal variables")),
        0x66: (PrinterError, _("No cliche")),
        0x47: (PrinterError, _("No company data. Has the printer been "
                               "initialized?")),
        0x67: (DriverError, "Invalid voucher amount or quantity"),
        0x48: (DriverError, ("Invalid managemental report number or "
                             "quantity")),
        0x68: (DriverError, "There is no more copies for the tied coupon"),
        0x49: (CommandError, _("Invalid command")),
        0x69: (CommandParametersError, _("Invalid command parameters")),
        0x4a: (DriverError, "Sale subjects to ICMS without state registry"),
        0x4d: (PrinterError, _('Fiscal memory without logotype')),
        0x6d: (HardwareFailure, _("Write error on the Fiscal Memory")),
        0x4e: (CommandError, _("Invalid state")),
        0x6e: (DriverError, "Invalid 'finalizadora' number"),
        0x50: (OutofPaperError, _("Printer is running out of paper")),
        0x70: (HardwareFailure, _("Printer hardware failure")),
        0x52: (PendingReduceZ, _("Pending Reduce Z")),
        0x53: (DriverError, ("Sale subjects to ISSQN without state "
                             "registry.")),
        0x73: (PrinterError, _("Discount in subtotal with sale subjects to "
                               "ICMS and ISSQN isn't allowed")),
        0x54: (DriverError, "Wrong tribute index or number"),
        0x74: (DriverError, "Found the 'TOTAL' word and/or its variables"),
        0x55: (DriverError, "Invalid measurement unit"),
        0x56: (DriverError, ("Total item value is greater than maximum "
                             "allowed")),
        0x76: (DriverError, "Attempt to cancel coupon at zero"),
        0x77: (DriverError, "Total item value is zero"),
        0x58: (PendingReadX, _("Pending Read X")),
        0x59: (ReduceZError, _("Attempt of reduce Z with date previous than "
                               "last")),
        0x79: (DriverError, ("Attempt of adjust the clock to date/time "
                             "previous than the last reduce Z")),
        0x7a: (HardwareFailure, _("No more fiscal memory :(")),
        0x5a: (ReduceZError, _("Reduce Z already done"))
    }

    def __init__(self, raw_status):
        self.parse(raw_status)

    def parse_error(self):
        status = ord(self.statuses)

        # There is no error to parse.
        if status == EP375Status.PRINTER_IS_OK:
            return
        try:
            exception, reason = self.errors_dict[status]
            raise exception(reason)
        except KeyError:
            raise DriverError("Unhandled error: %d" % status)

    def needs_reduce_Z(self):
        return ord(self.internal_state) == EP375Status.NEEDS_REDUCE_Z

    def needs_read_X(self):
        return ord(self.internal_state) == EP375Status.NEEDS_READ_X

    def has_been_totalized(self):
        return ord(self.internal_state) == EP375Status.HAS_BEEN_TOTALIZED

    def has_opened_sale(self):
        return (self.has_been_totalized()
                or ord(self.internal_state) in (EP375Status.HAS_NO_FISCAL_SALE,
                                                EP375Status.HAS_FISCAL_SALE))

    def has_opened_report(self):
        return ord(self.internal_state) == EP375Status.HAS_OPENED_REPORT

    def parse(self, status):
        """ This method parse the result of the 'GET_STATUS' command (see
        chart below).

        Status format::

          L S N N S K
          | | | | | |
          | | | | | +--> The warning (one of items in self.errors_dict)
          | | | | +----> Has CMC7? ('S'/'N')
          | | | +------> Opened Drawer? ('S'/'N')
          | | +--------> Technical mode? ('S'/'N')
          | +----------> Printer is ready? ('S'/'N')
          +------------> Internal state (page 15)
        """
        (self.internal_state,
         self.is_ready,
         self.technical_model,
         self.opened_drawer,
         self.has_cmc7,
         self.statuses) = status

    def __repr__(self):
        return ("<%s internal_state=%s is_ready=%r>" % (self.__class__.__name__,
                                                        self.internal_state,
                                                        self.is_ready))


class CouponItem:
    def __init__(self, code, description, taxcode, quantity, price, discount,
                 surcharge, unit):
        self.code = code
        self.description = description
        self.taxcode = taxcode
        self.quantity = quantity
        self.price = price
        self.discount = discount
        self.surcharge = surcharge
        self.unit = unit

    def get_packaged(self):
        if len(self.description) > 20:
            desc_size = 60
        else:
            desc_size = 20

        if self.discount:
            D = self.discount
        else:
            D = self.surcharge

        taxcode = self.taxcode
        quantity = int(float(self.quantity) * 1e3)
        price = int(float(self.price) * 1e2)
        D = int(float(D) * 1e2)

        return ("%-16s"  # code
                "%-*s"  # description
                "%02s"  # taxcode
                "%06d"  # quantity
                "%09d"  # price
                "%04d"  # discount/surcharge
                "%02s"  # unit
                % (self.code[:16], desc_size, self.description[:desc_size],
                   taxcode, quantity, price, D, self.unit))


#
# The driver implementation
#

class EP375(SerialBase, BaseChequePrinter):

    implements(ICouponPrinter, IChequePrinter)

    model_name = "Dataregis 375 EP"
    cheque_printer_charset = "ascii"
    coupon_printer_charset = "ascii"

    CHEQUE_CONFIGFILE = "dataregis.ini"

    CMD_PREFIX = '\xfe'

    #
    # IChequePrinter commands specification
    #

    CMD_CHEQUE = 'U'
    CMD_PRINT_CHEQUE = 'A'

    #
    # ICouponPrinter Commands specification
    #

    CMD_READ_X = 'G'
    CMD_REDUCE_Z = 'H'
    CMD_ADD_ITEM_WITH_DISCOUNT = 'A'
    CMD_ADD_ITEM_WITH_SURCHARGE = 'v'
    CMD_GET_STATUS = 'R'
    CMD_CANCEL_COUPON = 'F'
    CMD_ADD_PAYMENT = 'D'
    CMD_ADD_PAYMENT_WITH_SURCHARGE = 'c'
    CMD_CANCEL_ITEM = 'b'
    CMD_GET_REMAINING_VALUE = 'C'
    CMD_GET_FISCAL_COUNTERS = 'o'
    CMD_GERENCIAL_REPORT = 'j'
    CMD_CLOSE_GERENCIAL_REPORT = 'k'

    def __init__(self, port, consts):
        SerialBase.__init__(self, port)
        BaseChequePrinter.__init__(self)
        self._consts = consts or EP375Constants
        self.coupon_discount = Decimal("0.0")
        self.coupon_surcharge = Decimal("0.0")
        self._command_id = self._item_counter = -1
        self.items_dict = {}
        self._is_coupon_open = False

        # The printer needs a little delay to shutdown/startup. Add this
        # number as a guard to make sure that the printer doesn't freeze
        # if we start up too fast.
        time.sleep(0.2)

    #
    # Helper methods
    #

    def _get_next_command_id(self):
        self._command_id += 1
        return self._command_id

    def _get_next_coupon_item_id(self):
        self._item_counter += 1
        return self._item_counter

    def _unpack(self, package):
        if package[0] != EP375.CMD_PREFIX:
            raise ValueError("Received inconsistent data")
        n_params = ord(package[3])
        params = package[4:4 + n_params]
        if len(params) != n_params:
            raise ValueError("Received inconsistent data")
        return params

    def _is_valid_package(self, package):
        # minimum package size:
        #     STX|CMD_COUNTER|CMD_ID|PARAMS_SZ|CHECKSUM => 5 bytes
        if len(package) < 5:
            return False
        try:
            params = self._unpack(package)
        except ValueError:
            return False
        cmdid = package[2]
        checksum = ord(package[4 + len(params)])
        new_checksum = ord(self._get_packed(cmdid, params)[-1])
        return new_checksum == checksum

    def _get_and_parse_error(self):
        self._get_status().parse_error()

    def _parse_reply(self, reply):
        result = ""
        firstbyte = ord(reply[0])
        # When ACK+CR is received the command wasn't executed.  In this
        # case, we need call the printer status and manage the reason.
        if firstbyte == ACK:
            return self._get_and_parse_error()
        # When EOT is received, no reply is required (the printer will
        # send us nothing)
        elif firstbyte == EOT:
            return ""
        # When BS is received, the printer will send data yet, so we need
        # only executed the loop below
        elif firstbyte == BS:
            pass
        # Hmmm, if a broken package was sent we need manage this sending
        # ACK to the printer ("hey printer, you give me a broken package")
        # and so we need append the new reply in the current one and call
        # recursively this method.
        elif not self._is_valid_package(reply):
            self.write(chr(ACK))
            return self._parse_reply(reply + self.readline())
        else:
            result = self._unpack(reply)
        while True:
            lastbyte = ord(reply[-1])
            if lastbyte == SUB:
                break
            reply = self.readline()
            return result + self._parse_reply(reply)
        return result

    def _get_coupon_remaining_value(self):
        #
        # reply format:
        #     STX | CMD_COUNTER | CMD_ID | PARAMS_SZ | PARAMS | CHECKSUM
        # where PARAMS is:
        #     XYYYYYYYYYYYYYYZZZ
        # with:
        #     X == 'S' if the coupon is not completely paid
        #     Y == 14 bytes representing the remaining value
        #     Z == the number of items in the coupon.

        # We get the result already unpacked by send_command
        result = self._send_command(self.CMD_GET_REMAINING_VALUE)
        if result[0] == 'S':
            return Decimal(result[1:-3]) / Decimal("1e2")
        return Decimal("0.0")

    def _get_fiscal_counters(self):
        return self._send_command(self.CMD_GET_FISCAL_COUNTERS)

    def _get_coupon_number(self):
        result = self._get_fiscal_counters()
        return int(result[43:49])

    #
    # SerialBase wrappers
    #

    def writeline(self, data):
        while not self._port.getDSR():
            pass
        return SerialBase.writeline(self, data)

    def readline(self):
        self._port.setDTR()
        return SerialBase.readline(self)

    #
    # Methods implementation to printer commands and reply management
    #

    def _get_packed(self, command, *params):
        # Create a package for the command and its parameters. Package
        # format:
        #
        #         +---+-----------+------+---------+--------+--------+
        #NAME:    |STX|CMD_COUNTER|CMD_ID|PARAMS_SZ|[PARAMS]|CHECKSUM|
        #         +---+-----------+------+---------+--------+--------+
        #N_BYTES:   1       1         1        1        ?        1
        #
        # Where:
        #
        # CMD_COUNTER: the counter for the respective command sent (can be 0)
        # CMD_ID: the ID of the command
        # PARAMS_SZ: the number of parameters for the command
        # PARAMS: the params listing (it's optional and only is used if
        #   PARAMS_SZ is greater than 0)
        # CHECKSUM: the sum of all the bytes, starting at CMD_ID and ending
        #   at the last byte of PARAMS  (or PARAMS_SZ, if no parameters was
        #   sent, of course).
        params = ''.join(params)
        data = '%s%c%s' % (command, len(params), params)
        checksum = sum([ord(d) for d in data]) & 0xff
        package = '%c%s%c' % (self._get_next_command_id(),
                              data, checksum)
        return package

    def _send_command(self, command, *params):
        reply = self.writeline(self._get_packed(command, *params))
        result = self._parse_reply(reply)
        self.write(chr(EOT))
        return result

    def _get_status(self):
        result = self._send_command(self.CMD_GET_STATUS)
        return EP375Status(result)

    #
    # ICouponPrinter implementation
    #
    def coupon_identify_customer(self, customer, address, document):
        # The printer Dataregis 375-EP doesn't supports customer
        # identification
        return

    def coupon_is_customer_identified(self):
        return False

    def coupon_open(self):
        #
        # Dataregis 375-EP doesn't need a function to open a coupon - the
        # coupon is opened when the first item is added, so simple checks
        # is done at this part.
        #

        status = self._get_status()

        if status.needs_reduce_Z():
            raise PendingReduceZ(_("Pending Reduce Z"))
        if status.needs_read_X():
            raise PendingReadX(_("Pending Read X"))
        if status.has_been_totalized() or self._is_coupon_open:
            raise CouponOpenError("There is a coupon opened")
        self._is_coupon_open = True

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        if not self._is_coupon_open:
            raise CouponNotOpenError("There is no coupon opened")
        if unit == UnitType.CUSTOM:
            unit = UnitType.EMPTY
        if surcharge:
            cmd = self.CMD_ADD_ITEM_WITH_SURCHARGE
        else:
            cmd = self.CMD_ADD_ITEM_WITH_DISCOUNT

        #
        # FIXME: The product code can only contain alphanumeric characters if
        # the taxcode is between 90-99, i.e, if the product isn't tied to ICMS,
        # otherwise the printer will not recognizes the command. Sooooo, what
        # can I do? Right now, if the product code has not only numbers, i'll
        # prefix with 0s to avoid more problems to the callsite, but my warning
        # remains *HERE*
        #
        code_num = 0
        try:
            code_num = int(code[:7])
        except ValueError:
            code = "%06d%s" % (code_num, code[7:])
        unit = self._consts.get_value(unit)
        item = CouponItem(code, description, taxcode, quantity, price,
                          discount, surcharge, unit)
        item_id = self._get_next_coupon_item_id()
        self.items_dict[item_id] = item

        self._send_command(cmd, item.get_packaged())
        return item_id

    def coupon_cancel_item(self, item_id):
        try:
            item = self.items_dict[item_id]
        except KeyError:
            raise CancelItemError(_("You have specified an invalid item id "
                                    "to cancel!"))
        self._send_command(self.CMD_CANCEL_ITEM, item.get_packaged())

    def coupon_cancel(self):
        status = self._get_status()
        if status.has_opened_sale():
            self.coupon_add_payment(PaymentMethodType.MONEY,
                                    self._get_coupon_remaining_value())
        elif status.has_opened_report():
            self._send_command('K')
        # We can have the "coupon state flag" set to True, but no coupon really
        # opened; and we *must* to manage this case too...
        elif self._is_coupon_open:
            return
        else:
            raise CouponNotOpenError("There is no coupon opened")
        self._send_command(self.CMD_CANCEL_COUPON)
        self._is_coupon_open = False

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        self._send_command(self.CMD_CANCEL_COUPON)

    def coupon_totalize(self, discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), taxcode=TaxType.NONE):
        # FIXME: API changed: discount/surcharge was percentage,
        # now is currency.

        # The callsite must check if discount and charge are used together,
        # if so must raise an exception -- here we have a second check for
        # this.
        assert not (discount and surcharge)

        # Dataregis doesn't contains any functions to totalize the coupon,
        # so we need just save the discount/surcharge (if any) to use in the
        # add_payment calls
        coupon_subtotal = self._get_coupon_remaining_value()
        self.coupon_surcharge = coupon_subtotal * (surcharge / Decimal("100"))
        self.coupon_discount = coupon_subtotal * (discount / Decimal("100"))
        return coupon_subtotal + self.coupon_surcharge - self.coupon_discount

    def coupon_add_payment(self, payment_method, value, description='',
                           custom_pm=''):
        if not custom_pm:
            pm = self._consts.get_value(payment_method)
        else:
            pm = custom_pm
        value = "%014d" % int(float(value) * 1e2)

        if ((not self._get_status().has_been_totalized())
                and self.coupon_discount or self.coupon_surcharge):
            if self.coupon_discount:
                type = "D"
                D = self.coupon_discount
            else:
                type = "A"
                D = self.coupon_surcharge
            D = "%014d" % int(float(D) * 1e2)
            self._send_command(self.CMD_ADD_PAYMENT_WITH_SURCHARGE, pm, value,
                               D, type)
        else:
            self._send_command(self.CMD_ADD_PAYMENT, pm, value)
        return self._get_coupon_remaining_value()

    def coupon_close(self, message=''):
        # XXX: Here we have a problem -- what we can do with the 'message'?
        # Maybe this driver implementation requires a change in the package,
        # adding an extra function "set_coupon_footer_message" to the API.
        # With this function this driver and all the another ones will allow
        # the "promotional_message" in the coupon.
        self._is_coupon_open = False
        return self._get_coupon_number()

    def close_till(self, previous_day=False):
        if not self._get_status().needs_reduce_Z():
            raise ReduceZError(_('Reduce Z already done'))
        else:
            self._send_command(self.CMD_REDUCE_Z)

    def summarize(self):
        self._send_command(self.CMD_READ_X)

    def till_add_cash(self, value):
        value = format_value(value, 32)
        self._send_command(self.CMD_GERENCIAL_REPORT, "01", "Valor = " + value)
        self._send_command(self.CMD_CLOSE_GERENCIAL_REPORT)

    def till_remove_cash(self, value):
        value = format_value(value, 32)
        self._send_command(self.CMD_GERENCIAL_REPORT, "01", "Valor = " + value)
        self._send_command(self.CMD_CLOSE_GERENCIAL_REPORT)

    def till_read_memory(self, start, end):
        raise NotImplementedError

    def till_read_memory_by_reductions(self, start, end):
        raise NotImplementedError

    def get_capabilities(self):
        # FIXME: As always, we have a problem here with Dataregis printer:
        # only one of the last 100 items can be cancelled, so the 'item_id'
        # capability must have what value? Probably we never will have
        # more than 100 items right now, so I just put a mark here, this
        # must be fixed in the future.
        return dict(item_code=Capability(min_len=3, max_len=6),
                    item_id=Capability(digits=3),
                    items_quantity=Capability(min_size=1, digits=3, decimals=3),
                    item_price=Capability(digits=6, decimals=3),
                    item_description=Capability(max_len=60),
                    payment_value=Capability(digits=12, decimals=2),
                    promotional_message=Capability(),
                    payment_description=Capability(),
                    customer_name=Capability(),
                    customer_id=Capability(),
                    customer_address=Capability(),
                    cheque_thirdparty=Capability(max_len=50),
                    cheque_value=Capability(digits=12, decimals=2),
                    cheque_city=Capability(max_len=20),
                    add_cash_value=Capability(min_size=1, digits=30,
                                              decimals=2),
                    remove_cash_value=Capability(min_size=1, digits=30,
                                                 decimals=2))

    def get_constants(self):
        return self._consts

    #
    # IChequePrinter implementation
    #

    def send_cheque_command(self, command, *params):
        reply = self.writeline(self._get_packed(self.CMD_CHEQUE, command,
                                                *params))
        result = self._parse_reply(reply)
        # The printer is waiting for a 'End of Transmition' byte
        self.write(chr(EOT))
        return result

    def print_cheque(self, bank, value, thirdparty, city, date=None):
        if not isinstance(bank, BankConfiguration):
            raise TypeError("bank parameter must be a BankConfiguration "
                            "instance")
        if date is None:
            date = datetime.datetime.now()
        value = '%014d' % int(value * int(1e2))
        thirdparty = '%-50s' % thirdparty[:50]
        city = "%-20s" % city[:20]
        date = date.strftime("%d%m%y")

        positions = [bank.get_y_coordinate("value"),
                     bank.get_x_coordinate("value"),
                     bank.get_y_coordinate("legal_amount"),
                     bank.get_x_coordinate("legal_amount"),
                     bank.get_y_coordinate("legal_amount2"),
                     bank.get_x_coordinate("legal_amount2"),
                     bank.get_y_coordinate("thirdparty"),
                     bank.get_x_coordinate("thirdparty"),
                     bank.get_y_coordinate("city"),
                     bank.get_x_coordinate("city")]
        positions_data = "".join(["%02d" % pos for pos in positions])

        self.send_cheque_command(self.CMD_PRINT_CHEQUE, positions_data, value,
                                 thirdparty, city, date)



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Evandro Vale Miquelito      <evandro@async.com.br>
##              Henrique Romano             <henrique@async.com.br>
##              Johan Dahlin                <jdahlin@async.com.br>
##
"""
Stoqdrivers interfaces specification
"""

from decimal import Decimal

from kiwi.currency import currency
from zope.interface import Interface, Attribute

from stoqdrivers.enum import TaxType, UnitType

__all__ = ["ISerialPort",
           "IDriverConstants",
           "IDevice",
           "ICouponPrinter",
           "IChequePrinter",
           "IScaleInfo",
           "IScale",
           ]


class ISerialPort(Interface):
    """ Interface used by drivers to write commands and get reply from devices
    """

    def getDSR():
        """ Returns True if the device is done to send data. Some drivers
        block in a loop waiting for this function returns True before call
        read.
        """

    def setDTR(value):
        """ Set to True when the driver is going to send data to the device
        """

    def read(n_bytes=1):
        """ Read data """

    def write(data):
        """ Write data """


class IDriverConstants(Interface):
    """ This interface determines the methods to be implemented by all objects
    that wants didacte constant values for stoqdrivers devices drivers.
    """

    def get_items():
        """ Returns all the constant identifiers which this object has
        values assigned to.
        """

    def get_value(constant):
        """ Given one of the constants defined on stoqdrivers.constants,
        returns its value.
        """


class IDevice(Interface):
    model_name = Attribute("A string describing briefly the device implemented")


class ICouponPrinter(IDevice):
    """ Describes coupon related tasks for a printer.

    Workflow::

                                    --<--                     --<--
                                   |     |                   |     |
    [identify_customer] -> open -> add_item -> totalize -> add_payment -> close
    """

    coupon_printer_charset = Attribute("The charset name which the "
                                       "coupon printer uses.")

    supports_duplicate_receipt = Attribute(
        "Wheater or not the printer can print a "
        "duplicate payment receipt")

    identify_customer_at_end = Attribute(
        "True if the customer is identified at the"
        "end of the sale. False if at the beginning")

    #
    # Common API
    #

    def setup():
        """Setup printer for use.

        This method should do any logic required for the correct functioning of
        the printer, like checking if it uses 2 or 3 decimal places for product
        prices and quantity, and using this properties when necessary.
        """

    def coupon_identify_customer(customer, address, document):
        """ Identify the customer.  This method doesn't have mandatory
        execution (you can identify the customer only if you like), but when
        executed it must be called before calling any method.

        @param customer:
        @type customer:   str

        @param address:
        @type address:    str

        @param document:
        @type document:   str
        """

    def coupon_is_customer_identified():
        """ Returns True, if the customer have already been identified,
        False otherwise.
        """

    def has_open_coupon():
        """Returns True if the printer currently has an open coupon"""

    def coupon_open():
        """ This needs to be called before anything else (except
        identify_customer())
        """

    def coupon_add_item(code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        """ Adds an item to the coupon.

        @param code:      item code identifier
        @type  code:      str

        @param description:  description of product
        @type  desription: str

        @param price:     price
        @type  price:     Decimal

        @param taxcode:   constant to descrive the tax
        @type  taxcode:   str

        @param quantity:  quantity
        @type  quantity:  Decimal

        @param unit:      constant to describe the unit
        @type unit:       integer constant one of: UnitType.LITERS, UnitType.EMPTY,
                          UnitType.METERS, UnitType.WEIGHT, UnitType.CUSTOM.

        @param discount:  discount in %
        @type  discount:  Decimal between 0-100

        @param surcharge: surcharge in %
        @type  surcharge: Decimal between 0-100

        @param unit_desc: A 2-byte string representing the unit that applies to
                          the product.
        @type unit_desc:  str

        @rtype:           Decimal
        @returns:        identifier of added item
        """

    def coupon_cancel_item(item_id):
        """ Cancels an item, item_id must be a value returned by
        coupon_add_item

        @param item_id:   the item id
        """

    def coupon_cancel():
        """
        Cancels the currently open coupon or the last closed open.
        You cannot close a previous coupon if you already created
        a new one.
        """

    def cancel_last_coupon():
        """
        Cancel the last closed non-fiscal coupon or the last sale.
        """

    def coupon_totalize(discount=currency(0), surcharge=currency(0),
                        taxcode=TaxType.NONE):
        """ Closes the coupon applies addition a discount or surcharge and tax.
        This can only be called when the coupon is open, has items added and
        payments added.

        @param discount:  discount in value
        @type discount:   Value must be greater or equal than zero, and less
                          than coupon total value

        @param surcharge: surcharge in value
        @type  surcharge: Value must be greater or equal than zero

        @param taxcode:   constant to descrive the tax
        @type  taxcode:   integer constant one of: TaxType.NONE, TaxType.SUBSTITUTION,
                          TaxType.EXEMPTION

        @rtype:           Decimal
        @returns          the coupon total value
        """

    def coupon_add_payment(payment_method, value, description=u""):
        """
        Add payment method to coupon.
        @param payment_method: The payment method.
        @type payment_method:  a device specific value representing the payment value

        @param value:     The payment value
        @type value:      Decimal

        @param description: A simple description of the payment method to be
                            appended to the coupon.
        @type value:      unicode

        @rtype:           Decimal
        @returns:         the total remaining amount
        """

    def coupon_close(message=''):
        """ It needs to be possible to open new coupons after this is called.
        You must call coupon_totalize before calling this method.

        @param message:   promotional message
        @type message:    str

        @rtype:           int
        @returns:         identifier of the coupon.
        """

    #
    # Base admin operations
    #

    def summarize():
        """ Prints a summary of all sales of the day. In Brazil this is
        'read X' operation.
        """

    def open_till():
        """ Open the till so fiscal documents can be emitted.

        Some printers simply emit a read X document, but some printers may
        require an special command
        """

    def has_pending_reduce():
        """ Verify is has pending 'reduce z'
        """

    def close_till(previous_day=False):
        """ Close the till for the day, no other actions can be done after
        this is called. In Brazil this is 'reduce Z' operation

        @param previous_day: if this is to close a till opened previously
        """

    def till_add_cash(value):
        """ Add an till complement. This is called 'suprimento de caixa' on
        Brazil

        @param value:     The value added
        @type value:      Decimal
        """

    def till_remove_cash(value):
        """ Retire payments from the till. This is called 'sangria' on Brazil

        @param value:     The value to remove
        @type value:      Decimal
        """

    def till_read_memory(start, end):
        """
        Reads the fiscal memory, from the date start to the date end

        @param start: start date
        @type start: datetime.date
        @param end: end date
        @type end: datetime.date
        """

    def till_read_memory_by_reductions(start, end):
        """
        Reads the fiscal memory, from the start reductions to the end
        reductions

        @param start: start reductions
        @type start: int
        @param end: end reductions
        @type end: int
        """

    def get_payment_receipt_identifier(method_name):
        """If the printer requires an identifier to open a payment receipt,
        this method should return that identifier.

        @param method_name: get receipt for specific method name.
        """

    def payment_receipt_open(identifier, coo, method, value):
        """Opens a non-fiscal bound receipt for a payment

        @param identifier: a char (A-P) representing the receipt to use.
                           The receipt must already be configured at the printer.
        @param method: another char (A-P) for the payment method. Also, this must be
                       configured at the printer. The payment method must be bindable
        @param coo: coo for the coupon containing the payment
        @param value: the value of the payment
        """

    def payment_receipt_print(text):
        """Prints the payment receipt text

        @param text: text to be printed
        """

    def payment_receipt_close():
        """Closes previouly opened payment receipt
        """

    #
    # Getting printer status
    #

    def get_capabilities():
        """ Returns a capabilities dictionary, where the keys are the strings
        below and its values are Capability instances

        * item_code           (str)
        * item_id             (int)
        * items_quantity      (float)
        * item_price          (float)
        * item_description    (str)
        * payment_value       (float)
        * payment_description (str)
        * promotional_message (str)
        * customer_name       (str)
        * customer_id         (str)
        * customer_address    (str)
        * add_cash_value      (float)
        * remove_cash_value   (float)
        """

    def get_constants():
        """ Returns the object that implements IDriverConstants where the printer
        is going to get constant values from.
        """

    def get_payment_constants():
        """"""

    def get_serial():
        """Returns the serial number for the printer"""

    def get_sintegra():
        """
        Gets an object that implements ISintegraData or None.
        """

    def get_firmware_version():
        """Firmware version of the printer
        """

    def get_ccf():
        """Fiscal Coupon Counter

        @returns: the last document ccf
        @rtype: integer
        """

    def get_coo():
        """Operation Order Counter

        @returns: the last document coo
        @rtype: integer
        """

    def get_gnf():
        """Nonfiscal Operation General Counter

        @returns: gnf
        @rtype: integer
        """

    def get_crz():
        """Z Reduction Counter

        @returns: the last document crz
        @rtype: integer
        """


class IChequePrinter(IDevice):
    """ Interface specification for cheque printers. """

    cheque_printer_charset = Attribute("The charset name which the cheque "
                                       "printer uses.")

    def get_banks():
        """ Returns a dictionary of all banks supported by the printer. The
        dictionary's key is the bank name and its value are BankConfiguration
        instances (this classe [BankConfiguration] is used to store and manage
        the values of each section in the configuration file).
        """

    def print_cheque(bank, value, thirdparty, city, date=None):
        """ Prints a cheque

        @param bank:      the code of bank
        @type bank:       one of codes returned by get_banks method.

        @param value:     the value of the cheque
        @type value:      Decimal

        @param thirdparty: receiver of the cheque
        @type thirdparty: str

        @param city:
        @type city:       str

        @param date:      when the cheque was payed, optional
        @type date:       datetime
        """

    def get_capabilities():
        """ Returns a capabilities dictionary, where the keys are the strings
        below and its values are Capability instances

        * cheque_thirdparty   (str)
        * cheque_value        (Decimal)
        * cheque_city         (str)
        """


class INonFiscalPrinter(IDevice):
    """ Interface used to formatting texts in non fiscal printers.
    """

    def centralize():
        """ Centralize the text to be sent to coupon. """

    def descentralize():
        """ Descentralize the text to be sent to coupon. """

    def set_bold():
        """ The sent text will be appear in bold. """

    def unset_bold():
        """ Remove the bold option. """

    def print_line(data):
        """ Performs a line break to the given text. """

    def print_inline(data):
        """ Print a given text in a unique line. """

    def print_barcode(code):
        """ Print a barcode representing the given code. """

    def print_qrcode(code):
        """ Print a qrcode representing the given code. """


class IScaleInfo(Interface):
    """ This interface list the data read by the scale """
    weight = Attribute("The weight read")
    price_per_kg = Attribute("The KG read")
    total_price = Attribute("The total price. It is equivalent to "
                            "price_per_kg * weight")
    code = Attribute("The product code")


class IScale(IDevice):
    """ This interface describes how to interacts with scales.
    """

    def read_data():
        """ Read informations of the scale, returning an object
        that implements IScaleInfo interface.
        """


class IBarcodeReader(IDevice):
    """ Interface specification describing how to interacts with barcode
    readers.
    """

    def get_code():
        """ Returns the code read. Note that this function should be
        called only when there are data received (you can use
        notify_read() to be notified when data was received), or it
        will block in loop waiting the data.
        """


class ISintegraData(Interface):
    # This is used to generate a Sintegra 60M entry
    opening_date = Attribute('the day the till was opened')
    serial = Attribute('identifier of the printer')
    serial_id = Attribute('identifier of the printer, for the branch')
    coupon_start = Attribute('first coupon generated during the period')
    coupon_end = Attribute('last coupon generated during the period')
    crz = Attribute('Total number of times the till has been closed')
    cro = Attribute('Total number of times the till has been opened')
    period_total = Attribute('The value of all sales during the specified period')
    total = Attribute('The total value for all sales done by this printer')
    # This is used to generate a Sintegra 60A entry
    tax_total = Attribute('The total value including taxes')



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2011 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Stoq Team  <stoq-devel@async.com.br>
##
##
"""
A simple implementation of a virtual printer.
"""
import datetime
from decimal import Decimal
import json
import os

import gtk
import pango
from kiwi.python import Settable
from zope.interface import implements

from stoqdrivers.enum import PaymentMethodType, TaxType, UnitType
from stoqdrivers.exceptions import (CouponTotalizeError, PaymentAdditionError,
                                    CloseCouponError, CouponOpenError,
                                    CancelItemError, ItemAdditionError,
                                    DriverError, PrinterOfflineError)
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext


class CouponItem:
    def __init__(self, id, quantity, value):
        self.id, self.quantity, self.value = id, quantity, value

    def get_total_value(self):
        return self.quantity * self.value


class FakeConstants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: 'F',
        UnitType.METERS: 'G ',
        UnitType.LITERS: 'H',
        UnitType.EMPTY: 'I',
    }

    _payment_constants = {
        PaymentMethodType.MONEY: 'M',
        PaymentMethodType.CHECK: 'C',
        PaymentMethodType.BILL: 'B',
        PaymentMethodType.CREDIT_CARD: 'R',
        PaymentMethodType.DEBIT_CARD: 'D',
        PaymentMethodType.FINANCIAL: 'F',
        PaymentMethodType.GIFT_CERTIFICATE: 'G',
    }

    _payment_descriptions = {
        'M': 'Dinheiro',
        'C': 'Cheque',
        'B': 'Boleto',
        'R': u'Cartão Crédito',
        'D': u'Cartão Débito',
        'F': u'Financeira',
        'G': u'Vale Compras',
    }

    _tax_constants = [
        (TaxType.SUBSTITUTION, 'TS', None),
        (TaxType.EXEMPTION, 'TE', None),
        (TaxType.NONE, 'TN', None),
        (TaxType.CUSTOM, 'T1', Decimal(18)),
        (TaxType.CUSTOM, 'T2', Decimal(12)),
        (TaxType.CUSTOM, 'T3', Decimal(5)),
        (TaxType.SERVICE, 'S0', Decimal(3)),
    ]


class OutputWindow(gtk.Window):
    columns = 60

    def __init__(self, printer):
        self._printer = printer
        gtk.Window.__init__(self)
        self.set_title(_("ECF Emulator"))
        self.set_size_request(220, 320)
        self.move(0, 0)
        self.set_deletable(False)
        self.vbox = gtk.VBox(0, False)
        self.add(self.vbox)

        self._create_ui()

    def _create_ui(self):
        sw = gtk.ScrolledWindow()
        self.vbox.pack_start(sw, True, True)
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)

        self.textview = gtk.TextView()
        self.textview.modify_font(pango.FontDescription("Monospace 7"))
        sw.add(self.textview)
        self.buffer = self.textview.get_buffer()

        buttonbox = gtk.HBox()
        self.vbox.pack_start(buttonbox, False, False)

        self.b = gtk.ToggleButton(_("Turn off"))
        self.b.set_active(True)
        buttonbox.pack_start(self.b)
        self.b.connect("toggled", self._on_onoff__toggled)

    def _on_onoff__toggled(self, button):
        if button.get_active():
            self.b.set_label(_("Turn off"))
            self._printer.set_off(False)
        else:
            self.b.set_label(_("Turn on"))
            self._printer.set_off(True)

    def feed(self, text):
        self.buffer.props.text += text
        mark = self.buffer.get_insert()
        self.textview.scroll_mark_onscreen(mark)

    def feed_line(self, text=None):
        if not text:
            self.buffer.props.text += ('-' * self.columns) + '\n'
            return

        length = (self.columns - len(text) - 2) / 2
        self.buffer.props.text += '%s %s %s\n' % (
            '-' * length, text, '-' * length)
        mark = self.buffer.get_insert()
        self.textview.scroll_mark_onscreen(mark)


class Simple(object):
    implements(ICouponPrinter)

    model_name = "Virtual Printer"
    coupon_printer_charset = "utf-8"

    identify_customer_at_end = True
    supported = False

    def __init__(self, port, consts=None):
        self._consts = consts or FakeConstants()
        self._customer_document = None

        self._off = False
        self.output = OutputWindow(self)

        # Internal state
        self.till_closed = False
        self.opening_date = datetime.date.today()
        self.serial = 'Serial'
        self.serial_id = '1234567890'
        self.coupon_start = 0
        self.coupon_end = 10
        self.cro = 1
        self.crz = 1
        self.coo = 1
        self.gnf = 1
        self.ccf = 1
        self.period_total = 0
        self.total = 0
        self.taxes = []

        self._reset_flags()
        self._load_state()

        self.output.feed(
            "Virtual Printer\n"
            "Test Company\n"
            "CNPJ: 00.000.000/0000-00 IE: ISENTO\n"
            "IM: 012345\n")
        self.output.feed_line()
        self.output.feed("%s  COO:%s\n" % (
            datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
            self.coo))
        self.output.feed_line()

    #
    # Helper methods
    #

    def _get_state_filename(self):
        dirname = os.path.join(os.environ['HOME'], '.stoq')
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        filename = os.path.join(dirname, 'virtual-printer.json')
        return filename

    def _load_state(self):
        filename = self._get_state_filename()
        try:
            fp = open(filename, 'r')
        except (OSError, IOError):
            return
        try:
            state = json.load(fp)
        except ValueError:
            return

        self.till_closed = state['till-closed']

        fp.close()

    def _save_state(self):
        filename = self._get_state_filename()
        try:
            fp = open(filename, 'w')
        except (OSError, IOError):
            return

        state = {}
        state['till-closed'] = self.till_closed

        json.dump(state, fp)
        fp.write('\n')
        fp.close()

    def set_off(self, off):
        self._off = off

    def _check(self):
        if self._off:
            raise PrinterOfflineError

    def _reset_flags(self):
        self.is_coupon_opened = False
        self.items_quantity = 0
        self.is_coupon_totalized = False
        self.totalized_value = Decimal("0.0")
        self.has_payments = False
        self.payments_total = Decimal("0.0")
        self._items = {}

    def _check_coupon_is_opened(self):
        if not self.is_coupon_opened:
            raise CouponOpenError(_("There is no coupon opened!"))

    def _check_coupon_is_closed(self):
        if self.is_coupon_opened:
            raise CouponOpenError(_("There is a coupon already open"))

    #
    # ICouponPrinter implementation
    #

    def setup(self):
        pass

    def has_open_coupon(self):
        return self.is_coupon_opened

    def coupon_identify_customer(self, customer, address, document):
        self._check()
        if document not in ['160.618.061-40',
                            '335.728.854-00',
                            '804.727.615-87',
                            '871.007.004-42']:
            raise ItemAdditionError(
                _("Not allowed to sell to a client not created by the demo"))

        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return self._customer_document is not None

    def coupon_open(self):
        self._check()
        self.output.feed('\n')
        self.output.feed("CUPOM SIMULADO\n")

        self.output.feed("ITEM CODIGO DESCRICAO QTD.UN.VL. UNIT R$ ST A/T VL ITEM R$\n")
        self._check_coupon_is_closed()
        self.is_coupon_opened = True

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        self._check()
        if code not in ["{:02d}".format(x) for x in range(1, 11)]:
            # Allow deliveries as well
            if code != '' and taxcode != 'S0':
                raise ItemAdditionError(
                    _("Not allowed to sell an item not created by the demo"))

        self._check_coupon_is_opened()
        if self.is_coupon_totalized:
            raise ItemAdditionError(_("The coupon is already totalized, "
                                      "you can't add items anymore."))
        self.items_quantity += 1
        item_id = self.items_quantity
        item = CouponItem(item_id, quantity, price)
        self._items[item_id] = item
        self.output.feed("%03d %s %s\n" % (self.items_quantity, code, description))
        self.output.feed("  %d %f %s\n" % (quantity, price, taxcode))
        return item_id

    def coupon_cancel_item(self, item_id):
        self._check()
        self._check_coupon_is_opened()
        if not item_id in self._items:
            raise CancelItemError(_("There is no item with this ID (%d)")
                                  % item_id)
        elif self.is_coupon_totalized:
            raise CancelItemError(_("The coupon is already totalized, "
                                    "you can't cancel items anymore."))
        self._items.pop(item_id)
        self.output.feed('cancel_item %r\n' % (item_id, ))

    def coupon_cancel(self):
        self._check()
        # FIXME: If we don't have a coupon open, verify that
        #        we've opened at least one
        #self._check_coupon_is_opened()
        self.output.feed_line()
        self.output.feed('    Cupom Cancelado\n')
        self.output.feed_line()
        self._reset_flags()

    def coupon_totalize(self, discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), taxcode=TaxType.NONE):
        # FIXME: API changed: discount/surcharge was percentage,
        # now is currency.
        self._check()
        self._check_coupon_is_opened()
        if not self.items_quantity:
            raise CouponTotalizeError(_("The coupon can't be totalized, since "
                                        "there is no items added"))
        elif self.is_coupon_totalized:
            raise CouponTotalizeError(_("The coupon is already totalized"))

        for item_id, item in self._items.items():
            self.totalized_value += item.get_total_value()

        surcharge_value = self.totalized_value * surcharge / 100
        discount_value = self.totalized_value * discount / 100
        self.totalized_value += (
            Decimal(-discount_value).quantize(Decimal('.01')) +
            Decimal(surcharge_value).quantize(Decimal('.01')))

        if not self.totalized_value > 0:
            raise CouponTotalizeError(_("Coupon totalized must be greater "
                                        "than zero!"))

        self.is_coupon_totalized = True
        self.output.feed('\n')
        self.output.feed('Pagamentos:\n')
        return self.totalized_value

    def coupon_add_payment(self, payment_method, value, description=u"",
                           custom_pm=""):
        self._check()
        if not self.is_coupon_totalized:
            raise PaymentAdditionError(_("Isn't possible add payments to the "
                                         "coupon since it isn't totalized"))
        self.payments_total += value
        self.has_payments = True
        self.output.feed('  %s - %s\n' % (
            self._consts._payment_descriptions[payment_method], value))
        return self.totalized_value - self.payments_total

    def coupon_close(self, message=''):
        self._check()
        self._check_coupon_is_opened()
        if not self.is_coupon_totalized:
            raise CloseCouponError(_("Isn't possible close the coupon "
                                     "since it isn't totalized yet!"))
        elif not self.has_payments:
            raise CloseCouponError(_("Isn't possible close the coupon "
                                     "since there is no payments added."))
        elif self.totalized_value > self.payments_total:
            raise CloseCouponError(_("The payments total value doesn't "
                                     "match the totalized value."))
        troco = self.payments_total - self.totalized_value
        if troco:
            self.output.feed('Troco: %0.2f\n' % troco)
        self.output.feed_line()
        if message:
            self.output.feed(message)
        self.output.feed('\n')
        self._reset_flags()
        return 0

    def get_capabilities(self):
        self._check()
        # fake values
        return dict(item_code=Capability(max_len=48),
                    item_id=Capability(max_size=32767),
                    items_quantity=Capability(digits=14, decimals=4),
                    item_price=Capability(digits=14, decimals=4),
                    item_description=Capability(max_len=200),
                    payment_value=Capability(digits=14, decimals=4),
                    promotional_message=Capability(max_len=492),
                    payment_description=Capability(max_len=80),
                    customer_name=Capability(max_len=30),
                    customer_id=Capability(max_len=29),
                    customer_address=Capability(max_len=80),
                    cheque_thirdparty=Capability(max_len=45),
                    cheque_value=Capability(digits=14, decimals=4),
                    cheque_city=Capability(max_len=27))

    def get_constants(self):
        self._check()
        return self._consts

    def summarize(self):
        self._check()
        self.output.feed('LEITURA X\n')
        self.output.feed_line()
        self.till_closed = False
        self._save_state()

    def open_till(self):
        self.summarize()

    def close_till(self, previous_day=False):
        self._check()
        if self.till_closed:
            raise DriverError(
                "Reduce Z was already sent today, try again tomorrow")
        self.till_closed = True
        self._save_state()
        self.output.feed("REDUÇÃO Z\n")
        self.output.feed_line()

    def till_add_cash(self, value):
        self.output.feed('SUPRIMENTO: %s\n' % value)
        self.output.feed_line()
        self._check()

    def till_remove_cash(self, value):
        self.output.feed('SANGRIA: %s\n' % value)
        self.output.feed_line()
        self._check()

    def till_read_memory(self, start, end):
        self.output.feed('LEITURA MF\n')
        self.output.feed_line()
        self._check()

    def till_read_memory_by_reductions(self, start, end):
        self.output.feed('LEITURA MF\n')
        self.output.feed_line()
        self._check()

    def query_status(self):
        self._check()
        return None

    def get_serial(self):
        self.output.show_all()
        self._check()
        return 'Virtual'

    def get_tax_constants(self):
        self._check()
        return self._consts._tax_constants

    def get_payment_constant(self, payment):
        self._check()
        print 'payment', payment

    def get_payment_constants(self):
        self._check()
        return [(key, value) for key, value in
                self._consts._payment_descriptions.items()]

    def get_port(self):
        self._check()
        return None

    def has_pending_reduce(self):
        self._check()
        return False

    def get_coo(self):
        self._check()
        return self.coo

    def get_gnf(self):
        self._check()
        return self.gnf

    def get_ccf(self):
        self._check()
        return self.ccf

    def identify_customer_at_end(self):
        self._check()
        return False

    def get_sintegra(self):
        self._check()
        return Settable(opening_date=self.opening_date,
                        serial=self.serial,
                        serial_id=self.serial_id,
                        coupon_start=self.coupon_start,
                        coupon_end=self.coupon_end,
                        cro=self.cro,
                        crz=self.crz,
                        coo=self.coo,
                        period_total=self.period_total,
                        total=self.total,
                        taxes=self.taxes)

    def get_crz(self):
        return self.crz

    # Receipt

    def get_payment_receipt_identifier(self, method_name):
        self._check()
        return None

    def payment_receipt_open(self, identifier, coo, method, value):
        self.output.feed("    RECIBO DE PAGAMENTO coo=%s\n" % coo)

    def payment_receipt_print(self, text):
        self._check()
        self.output.feed(text)

    def payment_receipt_close(self):
        self._check()
        self.output.feed_line()

    def gerencial_report_open(self, gerencial_id=0):
        self._check()

    def gerencial_report_print(self, text):
        self._check()

    def gerencial_report_close(self):
        self._check()

    def get_firmware_version(self):
        return '1.1.1'

    #
    # IChequePrinter implementation
    #
    def print_cheque(self, value, thirdparty, city, date=None):
        return



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Henrique Romano             <henrique@async.com.br>
##
"""
Generic base class implementation for all printers
"""

from zope.interface import providedBy, implements
from kiwi.python import namedAny

from stoqdrivers.interfaces import (ICouponPrinter,
                                    IDriverConstants,
                                    IChequePrinter,
                                    INonFiscalPrinter)
from stoqdrivers.base import BaseDevice
from stoqdrivers.enum import DeviceType
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext

_NoDefault = object()


class BaseDriverConstants:
    implements(IDriverConstants)

    # Must be defined on subclasses
    _constants = None

    @classmethod
    def get_items(cls):
        return cls._constants.keys()

    @classmethod
    def get_value(cls, identifier, default=_NoDefault):
        try:
            return cls._constants[identifier]
        except KeyError:
            if default is not _NoDefault:
                return default
            raise ValueError("The constant identifier %r "
                             "isn't valid", identifier)


class BasePrinter(BaseDevice):
    device_dirname = "printers"
    device_type = DeviceType.PRINTER

    def check_interfaces(self):
        driver_interfaces = providedBy(self._driver)
        if (not ICouponPrinter in driver_interfaces
                and not IChequePrinter in driver_interfaces
                and not INonFiscalPrinter in driver_interfaces):
            raise TypeError("The driver `%r' doesn't implements a valid "
                            "interface" % self._driver)

    def get_constants(self):
        return self._driver.get_constants()

    def get_tax_constant(self, item):
        for enum, constant, value in self.get_tax_constants():
            if enum == item:
                return constant

    def get_model_name(self):
        return self._driver.model_name


def get_virtual_printer():
    from stoqdrivers.printers.fiscal import FiscalPrinter
    return FiscalPrinter(brand='virtual', model='Simple')


def get_supported_printers():
    result = {}
    for brand, module_names in [
            ('bematech', ['DP20C', 'MP20', 'MP2100', 'MP2100TH', 'MP25']),
            ('daruma', ['DR700', 'FS2100', 'FS345', 'FS600MFD']),
            ('dataregis', ['EP375', 'Quick']),
            ('elgin', ['KFiscal']),
            ('epson', ['FBII', 'FBIII']),
            ('fiscnet', ['FiscNetECF']),
            ('perto', ['Pay2023'])]:
        result[brand] = []
        for module_name in module_names:
            try:
                obj = namedAny("stoqdrivers.printers.%s.%s.%s"
                               % (brand, module_name, module_name))
            except AttributeError:
                raise ImportError("Can't find class %s for module %s"
                                  % (module_name, module_name))
            if not hasattr(obj, 'supported'):
                continue
            result[brand].append(obj)
    return result


def get_supported_printers_by_iface(interface):
    """ Returns all the printers that supports the interface.  The result
    format is the same for get_supported_printers."""
    if not interface in (ICouponPrinter, IChequePrinter, INonFiscalPrinter):
        raise TypeError("Interface specified (`%r') is not a valid "
                        "printer interface" % interface)
    all_printers_supported = get_supported_printers()
    result = {}
    for model, driver_list in all_printers_supported.items():
        drivers = []
        for driver in driver_list:
            if interface.implementedBy(driver):
                drivers.append(driver)
        if drivers:
            result[model] = drivers
    return result


def get_baudrate_values():
    """ Returns baudrate values to configure the communication speed with
    serial port.
    """
    return ['4800', '9600', '19200', '38400', '57600', '115200']



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2009 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Ronaldo Maia            <romaia@async.com.br>
##
"""
Dataregis Quick driver implementation.
"""

from stoqdrivers.printers.fiscnet.FiscNetECF import FiscNetECF
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext


class Quick(FiscNetECF):
    log_domain = 'DataregisQuick'
    supported = True

    model_name = "Dataregis ECF-IF 3202DT (Quick)"



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Evandro Vale Miquelito      <evandro@async.com.br>
##
"""
StoqDrivers exceptions definition
"""


class CriticalError(Exception):
    "Unknown device type or bad config settings"


class ConfigError(Exception):
    "Bad config file arguments or sections"


class CommError(Exception):
    "Common communication failures"


class PrinterError(Exception):
    "General printer errors"


class DriverError(Exception):
    "Base exception for all printer errors"
    def __init__(self, error='', code=-1):
        if code != -1:
            error = '%d: %s' % (code, error)
        Exception.__init__(self, error)
        self.code = code


class OutofPaperError(DriverError):
    "No paper left"


class PrinterOfflineError(DriverError):
    "Printer is offline"


class AlmostOutofPaper(DriverError):
    "Almost out of paper"


class HardwareFailure(DriverError):
    "Unknown hardware failure"


class PendingReduceZ(DriverError):
    "A Reduce Z is pending"


class PendingReadX(DriverError):
    "A Read X is pending"


class CloseCouponError(DriverError):
    "Could not close the coupon."


class CouponNotOpenError(DriverError):
    "Coupon is not open."


class CouponOpenError(DriverError):
    "Coupon already is open."


class AuthenticationFailure(DriverError):
    "General authentication failure"


class CommandParametersError(DriverError):
    "Parameters sent to printer are wrong."


class CommandError(DriverError):
    "Command sent to printer is wrong."


class ClosedTillError(DriverError):
    "No transactions can be done while the till is closed."


class ReduceZError(DriverError):
    "A Reduce already done."


class ReadXError(DriverError):
    "A Read X is already done."


class CouponTotalizeError(DriverError):
    "Error while totalizing a coupon."


class PaymentAdditionError(DriverError):
    "Error while adding a payment."


class CancelItemError(DriverError):
    "Error while cancelling coupon item."


class InvalidState(DriverError):
    "Invalid state for the requested operation."


class CapabilityError(Exception):
    "General capability error."


class ItemAdditionError(DriverError):
    "Error while adding an item."


class InvalidReply(DriverError):
    "Invalid reply received"


class AlreadyTotalized(DriverError):
    "The coupon is already totalized"


class InvalidValue(DriverError):
    "The value specified is invalid or is not in the expected range"



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):    Henrique Romano  <henrique@async.com.br>
##               Johan Dahlin     <jdahlin@async.com.br>
##

from ConfigParser import ConfigParser
import datetime

from zope.interface.exceptions import DoesNotImplement
from zope.interface import providedBy
from kiwi.argcheck import argcheck, number
from kiwi.environ import environ

from stoqdrivers.interfaces import IChequePrinter
from stoqdrivers.exceptions import ConfigError
from stoqdrivers.printers.base import BasePrinter
from stoqdrivers.utils import encode_text
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext


class BankConfiguration:
    """ This class store and manage the Cheque elements positions for a bank.
    """
    @argcheck(str, dict)
    def __init__(self, name, items):
        """ Create a new cheque configuration for the bank 'name'

        @param name:   The name of bank to which this configuration belongs to.
        @type name:    str
        @param items:  A dictionary where the key is the configuration name
                       and its values are the row and column coordinates,
                       respectively, eg: row['legal_amount1'] = (Y, X)
        @type items:   dict
        """
        self.name, self._items = name, items

    @argcheck(str)
    def get_coordinate(self, name):
        if not name in self._items:
            raise KeyError(name)
        return self._items[name]

    @argcheck(str)
    def get_x_coordinate(self, name):
        return self.get_coordinate(name)[1]

    @argcheck(str)
    def get_y_coordinate(self, name):
        return self.get_coordinate(name)[0]


class BaseChequePrinter:
    """ A base class for all printers that implements IChequePrinter interface.
    This class knows how to deal with the configuration file, that contains
    all the configuration of all supported cheques by the printer.

    @cvar CHEQUE_CONFIGFILE: This constant must be redefined in subclass and
                             must specify the filename where the cheque printer
                             configuration can be found.
    """
    CHEQUE_CONFIGFILE = None

    def __init__(self):
        self._banks = {}

    def get_banks(self):
        configfile = self.__module__.split('.')[-2] + '.ini'

        config = ConfigParser()
        filename = environ.get_resource_filename('stoqdrivers', 'conf',
                                                 configfile)
        if not config.read(filename):
            return None
        for section in config.sections():
            # With this, we'll have a dictionary in this format:
            # CONFIG_NAME: "Y,X"
            items = dict(config.items(section))
            try:
                bank = self._parse_bank(items)
            except ConfigError, errmsg:
                raise ConfigError("In section `%s' of `%s': %s"
                                  % (section, filename, errmsg))
            self._banks[int(section)] = bank
        return self._banks

    @argcheck(dict)
    def _parse_bank(self, items):
        if not 'name' in items:
            raise ConfigError("There is no bank name defined")
        coordinates = {}
        bank_name = items.pop('name')
        for name in ('value', 'legal_amount', 'legal_amount2', 'city',
                     'thirdparty', 'year', 'day', 'month'):
            if not name in items:
                raise ConfigError("The especification for `%s' was not found"
                                  % name)
            value = items.pop(name)
            if not ',' in value:
                raise ConfigError("Invalid value format for `%s'" % name)
            x, y = value.split(',')
            x = int(x or 0)
            y = int(y or 0)
            if x < 0 or y < 0:
                raise ConfigError("Negative values aren't allowed")
            coordinates[name] = (x, y)
        if items:
            raise ConfigError("The name(s) %s are invalid, busted "
                              "configuration" % ", ".join(items.keys()))
        return BankConfiguration(bank_name, coordinates)


class ChequePrinter(BasePrinter):
    def __init__(self, brand=None, model=None, device=None, config_file=None,
                 *args, **kwargs):
        BasePrinter.__init__(self, brand, model, device, config_file, *args,
                             **kwargs)
        if not IChequePrinter in providedBy(self._driver):
            raise DoesNotImplement("The driver %r doesn't implements the "
                                   "IChequePrinter interface" % self._driver)
        self._charset = self._driver.cheque_printer_charset

    def _format_text(self, text):
        return encode_text(text, self._charset)

    #
    # IChequePrinter interface
    #

    def get_banks(self):
        self.info("get_banks")
        return self._driver.get_banks()

    @argcheck(object, number, basestring, basestring, datetime.datetime)
    def print_cheque(self, bank, value, thirdparty, city, date=None):
        if date is None:
            date = datetime.datetime.now()
        self.info('print_cheque')
        return self._driver.print_cheque(bank, value,
                                         self._format_text(thirdparty),
                                         self._format_text(city), date)

    def get_capabilities(self):
        self.info("get_capabilities")
        return self._driver.get_capabilities()

#
# Testing
#


def test():
    printer = ChequePrinter()

    # Hmmm, this is not the right way of to do, but... hmm, I just don't know
    # what kind of the cheque the user wants to test, so let me get anyone.
    banks = printer.get_banks()
    bank = banks[banks.keys()[0]]
    printer.print_cheque(bank, 6.66, "Henrique Romano", u"São Paulo")

if __name__ == "__main__":
    test()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s): Henrique Romano             <henrique@async.com.br>
##
"""
Functions for general use.
"""

import unicodedata


def encode_text(text, encoding):
    """ Converts the string 'text' to encoding 'encoding' and optionally
    normalizes the string (currently only for ascii)

    @param text:       text to convert
    @type text:        str
    @param encoding:   encoding to use
    @type text:        str
    @returns:          converted text
    """
    if not isinstance(text, unicode):
        return text
    if encoding == "ascii":
        text = unicodedata.normalize("NFKD", text)
    return text.encode(encoding, "ignore")



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Johan Dahlin     <jdahlin@async.com.br>
##              Henrique Romano  <henrique@async.com.br>
##
"""
Daruma FS345 driver
"""
import datetime
from decimal import Decimal
import logging
import time

from kiwi.currency import currency
from kiwi.python import Settable
from zope.interface import implements

from stoqdrivers import abicomp
from stoqdrivers.serialbase import SerialBase
from stoqdrivers.interfaces import ICouponPrinter
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.printers.base import BaseDriverConstants
from stoqdrivers.enum import TaxType, UnitType
from stoqdrivers.exceptions import (DriverError, PendingReduceZ,
                                    HardwareFailure, ReduceZError,
                                    AuthenticationFailure, CouponNotOpenError,
                                    OutofPaperError, PrinterOfflineError,
                                    CouponOpenError, CancelItemError,
                                    CloseCouponError)
from stoqdrivers.translation import stoqdrivers_gettext

abicomp.register_codec()

_ = stoqdrivers_gettext

log = logging.getLogger('stoqdrivers.daruma')

CMD_STATUS = '\x1d\xff'

CMD_SET_OWNER = 190
# [ESC] 191 Gravação da indicação de mudança de moeda
# [ESC] 192 Intervenção Técnica
CMD_GET_MODEL = 195
CMD_GET_FIRMWARE = 199
CMD_OPEN_COUPON = 200
CMD_IDENTIFY_CUSTOMER = 201
CMD_ADD_ITEM_1L6D = 202
CMD_ADD_ITEM_2L6D = 203
CMD_ADD_ITEM_3L6D = 204
CMD_CANCEL_ITEM = 205
# [ESC] 206 Cancelamento de Documento
CMD_CANCEL_COUPON = 206
CMD_GET_X = 207
CMD_REDUCE_Z = 208
CMD_READ_MEMORY = 209
# [ESC] 210 Emissão de Cupom Adicional
# [ESC] 211 Abertura de Relatório Gerencial (Leitura X)
CMD_GERENCIAL_REPORT_OPEN = 211
CMD_GERENCIAL_REPORT_CLOSE = 212
CMD_GERENCIAL_REPORT_PRINT = 213
CMD_CLOSE_NON_FISCAL_BOUND_RECEIPT = 212        # Pg. 37
CMD_PRINT_LINE_NON_FISCAL_BOUND_RECEIPT = 213   # Pg. 37
CMD_ADD_ITEM_1L13D = 214
CMD_ADD_ITEM_2L13D = 215
CMD_ADD_ITEM_3L13D = 216
CMD_OPEN_VOUCHER = 217
CMD_DESCRIBE_MESSAGES = 218
CMD_OPEN_NON_FISCAL_BOUND_RECEIPT = 219  # Pg 36
CMD_CONFIGURE_TAXES = 220
CMD_LAST_RECORD = 221
# [ESC] 223 Descrição de produto em 3 linhas com código de 13 dígitos
#           (Formato fixo p/ Quantidade 5,3)
CMD_ADD_ITEM_3L13D53U = 223
# [ESC] 225 Descrição de produto com preço unitário com 3 decimais
CMD_DESCRIBE_NON_FISCAL_RECEIPT = 226   # Pg. 42
# [ESC] 227 Subtotalização de Cupom Fiscal
# [ESC] 228 Configuração da IF
CMD_GET_CONFIGURATION = 229
# [ESC] 230 Leitura do relógio interno da impressora
CMD_GET_TAX_CODES = 231
# [ESC] 232 Leitura do clichê do proprietário
# [ESC] 234 Retransmissão de mensagens da IF
CMD_GET_IDENTIFIER = 236
CMD_GET_PERSONAL_MESSAGES = 238
CMD_GET_DOCUMENT_STATUS = 239
CMD_GET_FISCAL_REGISTRIES = 240
CMD_TOTALIZE_COUPON = 241
CMD_DESCRIBE_PAYMENT_FORM = 242
CMD_CLOSE_COUPON = 243
CMD_GET_REGISTRIES = 244
CMD_GET_TOTALIZERS = 244
# [ESC] 246 Leitura Horária
# [ESC] 247 Descrição Estendida
# [ESC] 248 Estorno de forma de pagamento
CMD_GET_DATES = 250
# [ESC] 251 Leitura das informações cadastrais do usuário
# [ESC] V   Controle de horário de verão

CASH_IN_TYPE = 'B'
CASH_OUT_TYPE = 'A'

RETRIES_BEFORE_TIMEOUT = 5

# Document status
OPENED_FISCAL_COUPON = '1'
CLOSED_COUPON = '2'


def isbitset(value, bit):
    # BCD crap
    return (int(value, 16) >> bit) & 1 == 1


def ifset(value, bit, false='', true=''):
    if not isbitset(value, bit):
        return false
    else:
        return true


class FS345Constants(BaseDriverConstants):
    _constants = {
        UnitType.WEIGHT: 'Kg',
        UnitType.METERS: 'm ',
        UnitType.LITERS: 'Lt',
        UnitType.EMPTY: '  ',
    }


class FS345(SerialBase):
    log_domain = 'fs345'

    implements(ICouponPrinter)

    supported = True
    model_name = "Daruma FS 345"
    coupon_printer_charset = "abicomp"
    supports_duplicate_receipt = False
    identify_customer_at_end = True

    def __init__(self, port, consts=None):
        self._consts = consts or FS345Constants
        SerialBase.__init__(self, port)
        self._reset()

    def _reset(self):
        self._customer_name = u""
        self._customer_document = u""
        self._customer_address = u""

    def send_command(self, command, extra=''):
        raw = chr(command) + extra
        while True:
            self.write(self.CMD_PREFIX + raw + self.CMD_SUFFIX)
            retval = self._read_reply()
            if retval.startswith(':E'):
                try:
                    self.handle_error(retval, raw)
                except DriverError, e:
                    if e.code == 42:
                        self.send_command(CMD_GET_X)
                        continue
                    raise
            break

        return retval[1:]

    def _read_reply(self):
        rep = ''
        timeouts = 0

        while True:
            if timeouts > RETRIES_BEFORE_TIMEOUT:
                raise DriverError(_("Timeout communicating with fiscal "
                                    "printer"))

            c = self.read(1)
            if len(c) != 1:
                timeouts += 1
                continue

            if c == self.EOL_DELIMIT:
                log.debug("<<< %r" % rep)
                return rep

            rep += c

    # Status
    def _get_status(self):
        self.write(CMD_STATUS)
        return self._read_reply()

    def status_check(self, S, byte, bit):
        return isbitset(S[byte], bit)

    def _check_status(self, verbose=False):
        status = self._get_status()
        if status[0] != ':':
            raise HardwareFailure('Broken status reply')

        if verbose:
            print '== STATUS =='

            # Codes found on page 57-59
            print 'Raw status code:', status
            print 'Cashier drawer is', ifset(status[1], 3, 'closed', 'open')

        if self.has_pending_reduce():
            raise PendingReduceZ(_('Pending Reduce Z'))
        if self.status_check(status, 1, 2):
            raise HardwareFailure(_('Mechanical failure'))
        if not self.status_check(status, 1, 1):
            raise AuthenticationFailure(_('Not properly authenticated'))
        if self.status_check(status, 1, 0):
            raise OutofPaperError(_('No paper'))
        if self.status_check(status, 2, 3):
            raise PrinterOfflineError(_("Offline"))
        #if not self.status_check(status, 2, 2):
        #     raise CommError(_("Peripheral is not connected to AUX"))
        if self.status_check(status, 2, 0):
            log.info('Almost out of paper')

        if verbose:
            S3 = status[3]
            print ifset(S3, 3, 'Maintenance', 'Operational'), 'mode'
            print 'Authentication', ifset(S3, 2, 'disabled', 'enabled')
            print 'Guillotine', ifset(S3, 1, 'disabled', 'enabled')
            print 'Auto close CF?', ifset(S3, 0, 'no', 'yes')

        if self.status_check(status, 6, 1):
            raise ReduceZError(_("readZ is already emitted"))

        # FIXME: I am not sure we should be doing this here. This method
        # should only check the status, and not emit any other command.
        #if self.needs_read_x(status):
        #    self.send_command(CMD_GET_X)

        return status

    def has_pending_reduce(self, status=None):
        if not status:
            status = self._get_status()
        return self.status_check(status, 2, 1)

    def needs_read_x(self, status=None):
        if not status:
            status = self._get_status()

        return not self.status_check(status, 6, 2)

    # Error handling

    def handle_error(self, error_value, raw):
        error = int(error_value[2:])
        # Page 61-62
        if error == 39:
            raise DriverError('Bad parameters: %r' % raw, error)
        elif error == 10:
            raise CouponOpenError(_("Document is already open"), error)
        elif error == 11:
            raise CouponNotOpenError(_("Coupon is not open"), error)
        elif error == 12:
            raise CouponNotOpenError(_("There's no open document to cancel"),
                                     error)
        elif error == 15:
            raise CancelItemError(_("There is no such item in "
                                    "the coupon"), error)
        elif error == 16:
            raise DriverError("Bad discount/markup parameter", error)
        elif error == 21:
            log.warning(_('Printer error %s: No paper'), error)
        elif error == 22:
            raise DriverError(
                "Reduce Z was already sent today, try again tomorrow", error)
        elif error == 23:
            raise PendingReduceZ
        elif error == 24:
            raise DriverError("Bad unit specified: %r" % raw, error)
        elif error == 42:
            raise DriverError("Read X has not been sent yet", error)
        elif error == 45:
            raise DriverError(_("Required field is blank"), error)
        else:
            raise DriverError("Unhandled error: %d" % error, error)

    # Information / debugging

    def show_status(self):
        self._check_status(verbose=True)

    def show_information(self):
        print 'Model:', self.send_command(CMD_GET_MODEL)
        print 'Firmware:', self.send_command(CMD_GET_FIRMWARE)
        data = self.send_command(CMD_LAST_RECORD)

        tt = time.strptime(data[:12], '%d%m%y%H%M%S')
        print 'Last record:', time.strftime('%c', tt)
        print 'Configuration:', self.send_command(CMD_GET_CONFIGURATION)

    def show_document_status(self):
        print '== DOCUMENT STATUS =='
        value = self._get_document_status()
        print 'ECF:', value[2:6]
        document_type = value[6]
        if document_type == CLOSED_COUPON:
            print 'No open coupon'
        elif document_type == OPENED_FISCAL_COUPON:
            print 'Document is a coupon (%s)' % value[7:12]
        else:
            print 'Document type:', value[6]

        tt = time.strptime(value[13:27], '%H%M%S%d%m%Y')
        print 'Current date/time:', time.strftime('%c', tt)

        print 'Sum', int(value[27:41]) / 100.0
        print 'GT atual', value[41:59]

    def _get_document_status(self):
        status = self.send_command(CMD_GET_DOCUMENT_STATUS)
        assert status[:2] == '\x1b' + chr(CMD_GET_DOCUMENT_STATUS)
        assert len(status) == 59
        return status

    def get_firmware_version(self):
        """Return the firmware version."""
        return self.send_command(CMD_GET_FIRMWARE)

    def _get_fiscal_registers(self):
        value = self.send_command(CMD_GET_FISCAL_REGISTRIES)
        assert value[:2] == '\x1b' + chr(CMD_GET_FISCAL_REGISTRIES)
        return value[2:]

    def _get_registers(self):
        value = self.send_command(CMD_GET_REGISTRIES)
        assert value[:2] == '\x1b' + chr(CMD_GET_REGISTRIES)
        return value[2:]

    # High level commands
    def _verify_coupon_open(self):
        if not self.status_check(self._get_status(), 4, 2):
            raise CouponNotOpenError(_("Coupon is not open"))

    def _is_open(self, status):
        return self.status_check(status, 4, 2)

    # Helper commands

    def _get_totalizers(self):
        return self.send_command(CMD_GET_TOTALIZERS)

    def _get_coupon_number(self):
        return int(self._get_totalizers()[8:14])

    def _add_payment(self, payment_method, value, description=''):
        rv = self.send_command(CMD_DESCRIBE_PAYMENT_FORM,
                               '%c%012d%s\xff' % (payment_method,
                                                  int(value * Decimal('1e2')),
                                                  description[:48]))
        # FIXME: Why and when does this happen?
        #        Avoids/Fixes bug 3467 at least
        if rv[0] == 'N':
            rv = rv[8:]
        return Decimal(rv) / Decimal('1e2')

    def _add_voucher(self, type, value):
        data = "%s1%s%012d\xff" % (type, "0" * 12,  # padding
                                   int(value * Decimal('1e2')))
        self.send_command(CMD_OPEN_VOUCHER, data)

    def _configure_taxes(self):
        self.send_command(CMD_CONFIGURE_TAXES, '1800')
        self.send_command(CMD_CONFIGURE_TAXES, '1500')
        self.send_command(CMD_CONFIGURE_TAXES, '2500')
        self.send_command(CMD_CONFIGURE_TAXES, '0800')
        self.send_command(CMD_CONFIGURE_TAXES, '0500')
        self.send_command(CMD_CONFIGURE_TAXES, '0327')
        self.send_command(CMD_CONFIGURE_TAXES, '0592')
        self.send_command(CMD_CONFIGURE_TAXES, 'S0200')
        self.send_command(CMD_CONFIGURE_TAXES, 'S0300')
        self.send_command(CMD_CONFIGURE_TAXES, 'S0400')

    def _configure_payment_methods(self):
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGXADinheiro         ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGXBCheque           ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGXCBoleto           ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGVDCartao Credito   ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGVECartao Debito    ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGVFFinanceira       ')
        self.send_command(CMD_DESCRIBE_MESSAGES, 'PGVGVale Compra      ')

    def _configure_bound_receipts(self):
        self.send_command(CMD_DESCRIBE_NON_FISCAL_RECEIPT, 'VCartao Credito       ')

    #
    # API implementation
    #

    def setup(self):
        pass

    def coupon_identify_customer(self, customer, address, document):
        self._customer_name = customer
        self._customer_document = document
        self._customer_address = address

    def coupon_is_customer_identified(self):
        return len(self._customer_document) > 0

    def has_open_coupon(self):
        status = self._get_document_status()
        coupon_status = status[6]
        return coupon_status != CLOSED_COUPON

    def coupon_open(self):
        status = self._check_status()
        if self._is_open(status):
            raise CouponOpenError(_("Coupon already open"))
        self.send_command(CMD_OPEN_COUPON)

    def coupon_add_item(self, code, description, price, taxcode,
                        quantity=Decimal("1.0"), unit=UnitType.EMPTY,
                        discount=Decimal("0.0"),
                        surcharge=Decimal("0.0"), unit_desc=""):
        if surcharge:
            d = 1
            E = surcharge
        else:
            d = 0
            E = discount

        if unit == UnitType.CUSTOM:
            unit = unit_desc
        else:
            unit = self._consts.get_value(unit)

        if not code:
            code = "-"
        data = '%2s%13s%d%04d%010d%08d%s%s\xff' % (taxcode, code[:13], d,
                                                   int(E * Decimal('1e2')),
                                                   int(price * Decimal('1e3')),
                                                   int(quantity * Decimal('1e3')),
                                                   unit, description[:174])
        value = self.send_command(CMD_ADD_ITEM_3L13D53U, data)
        return int(value[1:4])

    def coupon_cancel_item(self, item_id):
        self.send_command(CMD_CANCEL_ITEM, "%03d" % item_id)

    def coupon_add_payment(self, payment_method, value, description=''):
        self._check_status()
        self._verify_coupon_open()
        return self._add_payment(payment_method, value, description)

    def coupon_cancel(self):
        # If we need reduce Z don't verify that the coupon is open, instead
        # just cancel the coupon. This avoids a race when you forgot
        # to close a coupon and reduce Z at the same time.
        if not self.has_pending_reduce():
            self._check_status()
        self.send_command(CMD_CANCEL_COUPON)

    def cancel_last_coupon(self):
        """Cancel the last non fiscal coupon or the last sale."""
        self.send_command(CMD_CANCEL_COUPON)

    def coupon_totalize(self, discount=currency(0), surcharge=currency(0),
                        taxcode=TaxType.NONE):
        self._check_status()
        self._verify_coupon_open()
        if surcharge:
            value = surcharge
            if taxcode == TaxType.ICMS:
                mode = 3
            else:
                raise ValueError("tax_code must be TaxType.ICMS")
        elif discount:
            value = discount
            mode = 1
        else:
            mode = 1
            value = currency(0)
        # Page 33
        data = '%s%012d' % (mode, int(value * Decimal("1e2")))
        rv = self.send_command(CMD_TOTALIZE_COUPON, data)
        return currency(rv) / Decimal("1e2")

    def coupon_close(self, message=''):
        self._check_status()
        self._verify_coupon_open()

        if (self._customer_name or
                self._customer_address or
                self._customer_document):
            customer_name = self._customer_name or _("No client")
            customer_document = self._customer_document or _("No document")
            customer_address = self._customer_address or _("No address")
            self.send_command(CMD_IDENTIFY_CUSTOMER,
                              "%- 84s%- 84s%- 84s" % (customer_name,
                                                      customer_address,
                                                      customer_document))
        message = message.encode(self.coupon_printer_charset)
        try:
            self.send_command(CMD_CLOSE_COUPON, message + '\xff')
        except DriverError:
            raise CloseCouponError(_("It is not possible to close the "
                                     "coupon"))
        self._reset()
        return self._get_coupon_number()

    def summarize(self):
        self.send_command(CMD_GET_X)

    def open_till(self):
        self.summarize()

    def close_till(self, previous_day=False):
        status = self._get_status()
        if self._is_open(status):
            self.send_command(CMD_CANCEL_COUPON)

        date = time.strftime('%d%m%y%H%M%S', time.localtime())
        self.send_command(CMD_REDUCE_Z, date)

    def _get_bound_receipt_constants(self):
        # Also page 48
        messages = self.send_command(CMD_GET_PERSONAL_MESSAGES)

        raw = messages[372:708]
        constants = []
        const_letter = 'ABCDEFGHIJKLMNOP'
        for i, char in enumerate(const_letter):
            const = raw[i * 21:i * 21 + 21]
            # Ignore constants that are not registered.
            if const[2] == '\xff':
                continue

            const = const.decode(self.coupon_printer_charset)
            constants.append((char, const.strip()))

        return constants

    def get_payment_receipt_identifier(self, method_name):
        """Returns the receipt identifier corresponding to the payment method.

        @param method_name: this is the payment method name. A receipt with
                            the same name should be configured at the printer.
        """
        constants = self._get_bound_receipt_constants()
        for id, name in constants:
            if name == method_name:
                return id

        raise DriverError(_("Receipt for method %s is not configured") % method_name)

    def payment_receipt_open(self, identifier, coo, method, value):
        value = int(value * 100)
        # res is the coo for the current document.
        self.send_command(CMD_OPEN_NON_FISCAL_BOUND_RECEIPT,
                          '%c%c%06d%012d' % (identifier, method, coo, value))

    def payment_receipt_print(self, text):
        for line in text.split('\n'):
            self.send_command(CMD_PRINT_LINE_NON_FISCAL_BOUND_RECEIPT,
                              line + chr(255))

    def payment_receipt_close(self):
        self.send_command(CMD_CLOSE_NON_FISCAL_BOUND_RECEIPT)

    def gerencial_report_open(self):
        self.send_command(CMD_GERENCIAL_REPORT_OPEN)

    def gerencial_report_print(self, text):
        for line in text.split('\n'):
            self.send_command(CMD_GERENCIAL_REPORT_PRINT, line + chr(255))

    def gerencial_report_close(self):
        self.send_command(CMD_GERENCIAL_REPORT_CLOSE)

    def till_add_cash(self, value):
        self._add_voucher(CASH_IN_TYPE, value)
        self._add_payment('A', value, '')

    def till_remove_cash(self, value):
        self._add_voucher(CASH_OUT_TYPE, value)

    def till_read_memory(self, start, end):
        # Page 39
        self.send_command(CMD_READ_MEMORY, 'x%s%s' % (start.strftime('%d%m%y'),
                                                      end.strftime('%d%m%y')))

    def _till_read_memory_to_serial(self, start, end):
        # Page 39
        ret = self.send_command(CMD_READ_MEMORY, 's%s%s' % (start.strftime('%d%m%y'),
                                                            end.strftime('%d%m%y')))
        data = ''
        while True:
            line = self.readline()
            if line[-1] == '\xff':
                break
            data += line

        data = unicode(data, 'cp860')
        return (ret, data)

    def till_read_memory_by_reductions(self, start, end):
        # Page 39
        self.send_command(CMD_READ_MEMORY, 'x00%04d00%04d' % (start, end))

    def get_capabilities(self):
        return dict(item_code=Capability(max_len=13),
                    item_id=Capability(digits=3),
                    items_quantity=Capability(min_size=1, digits=5, decimals=3),
                    item_price=Capability(min_size=0, digits=7, decimals=3),
                    item_description=Capability(max_len=173),
                    payment_value=Capability(digits=10, decimals=2),
                    promotional_message=Capability(max_len=384),
                    payment_description=Capability(max_len=48),
                    customer_name=Capability(max_len=42),
                    customer_id=Capability(max_len=42),
                    customer_address=Capability(max_len=42),
                    remove_cash_value=Capability(min_size=1, digits=10,
                                                 decimals=2),
                    add_cash_value=Capability(min_size=1, digits=10,
                                              decimals=2))

    def get_constants(self):
        return self._consts

    def query_status(self):
        return CMD_STATUS

    def status_reply_complete(self, reply):
        return '\r' in reply

    def get_serial(self):
        identifier = self.send_command(CMD_GET_IDENTIFIER)
        return identifier[1:9]

    def get_ccf(self):
        # Daruma FS345 does not have ccf. See:
        # http://www.forumweb.com.br/foruns/lofiversion/index.php/t64417.html
        return self.get_coo()

    def get_coo(self):
        registries = self._get_registers()
        return int(registries[7:12])

    def get_gnf(self):
        registries = self._get_registers()
        return int(registries[13:18])

    def get_crz(self):
        registries = self._get_registers()
        return int(registries[39:42])

    def get_tax_constants(self):
        tax_codes = self.send_command(CMD_GET_TAX_CODES)[1:]

        constants = []
        for i in range(14):
            reg = tax_codes[i * 5]
            if reg in 'ABCDEFGHIJKLMNOP':
                tax_type = TaxType.CUSTOM
            elif reg in 'abcdefghijklmnop':
                tax_type = TaxType.SERVICE
            else:
                raise AssertionError(reg)
            value = tax_codes[i * 5 + 1:i * 5 + 5]
            if value == '////':
                continue
            constants.append((tax_type,
                              'T' + reg.lower(),
                              Decimal(value.replace('.', '')) / 100))

        # These definitions can be found on Page 60
        constants.extend([
            (TaxType.SUBSTITUTION, 'Fb', None),
            (TaxType.EXEMPTION, 'Ib', None),
            (TaxType.NONE, 'Nb', None),
        ])

        return constants

    def get_payment_constants(self):
        # Page 48
        messages = self.send_command(CMD_GET_PERSONAL_MESSAGES)

        raw = messages[708:]
        methods = []
        method_letter = 'ABCDEFGHIJKLMNOP'
        for i in range(16):
            method = raw[i * 18:i * 18 + 18]
            # XXX V = vinculavel. Why Only These????
            #if method[0] == 'V':
            #    methods.append((method_letter[i], method[1:].strip()))
            if method[2] == '\xff':
                continue

            name = method[1:].strip().decode(self.coupon_printer_charset)
            methods.append((method_letter[i], name))

        return methods

    def get_sintegra(self):
        registries = self._get_registers()
        fiscal_registries = self._get_fiscal_registers()

        tax_codes = self.send_command(CMD_GET_TAX_CODES)[1:]

        taxes = []
        for i in range(14):
            reg = tax_codes[i * 5 + 1:i * 5 + 5]
            if reg == '////':
                continue
            reg = reg.replace('.', '')

            if tax_codes[i * 5] in 'ABCDEFGHIJKLMNOP':
                type = 'ICMS'
            else:
                type = 'ISS'

            sold = fiscal_registries[88 + (i * 14):102 + (i * 14)]
            taxes.append((reg, Decimal(sold) / 100, type))

        taxes.append(('DESC', Decimal(fiscal_registries[19:32]) / 100, 'ICMS'))
        taxes.append(('CANC', Decimal(fiscal_registries[33:46]) / 100, 'ICMS'))
        taxes.append(('I', Decimal(fiscal_registries[47:60]) / 100, 'ICMS'))
        taxes.append(('N', Decimal(fiscal_registries[61:74]) / 100, 'ICMS'))
        taxes.append(('F', Decimal(fiscal_registries[75:88]) / 100, 'ICMS'))

        total_sold = sum(value for _, value, _ in taxes)

        old_total = Decimal(fiscal_registries[:18]) / 100
        period_total = total_sold

        dates = self.send_command(CMD_GET_DATES)
        if dates[:6] == '000000':
            opening_date = datetime.date.today()
        else:
            d, m, y = map(int, [dates[:2], dates[2:4], dates[4:6]])
            opening_date = datetime.date(2000 + y, m, d)

        identifier = self.send_command(CMD_GET_IDENTIFIER)
        return Settable(
            opening_date=opening_date,
            serial=identifier[1:9],
            serial_id=int(identifier[13:17]),
            coupon_start=int(registries[:6]),
            coupon_end=int(registries[7:12]),
            cro=int(registries[35:38]),
            crz=int(registries[39:42]),  # FIXME: this is being fetched before the actual
            # reduction, so it will offset by one.
            coo=int(registries[6:12]),
            period_total=period_total,
            total=period_total + old_total,
            taxes=taxes)



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2006 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##
"""
Bematech DP20C driver
"""

import datetime

from zope.interface import implements

from stoqdrivers.interfaces import IChequePrinter
from stoqdrivers.printers.cheque import BaseChequePrinter
from stoqdrivers.printers.capabilities import Capability
from stoqdrivers.serialbase import SerialBase


class DP20C(SerialBase, BaseChequePrinter):
    CMD_PREFIX = '\x1B'
    CMD_SUFFIX = '\x0D'
    CMD_SETUP_COORDINATES = '\xAA'

    implements(IChequePrinter)

    model_name = "Bematech DP20C"
    cheque_printer_charset = "cp850"

    def __init__(self, port, consts=None):
        SerialBase.__init__(self, port)
        BaseChequePrinter.__init__(self)

    def _setup_positions(self, bank):
        # man page 24.
        data = [(bank.get_x_coordinate("value") - 60) / 10,
                (bank.get_x_coordinate("value") - 60) % 10,
                bank.get_y_coordinate("value"),
                bank.get_x_coordinate("legal_amount") / 10,
                bank.get_x_coordinate("legal_amount") % 10,
                bank.get_y_coordinate("legal_amount"),
                bank.get_x_coordinate("legal_amount2") % 10,
                bank.get_y_coordinate("legal_amount2") % 10,
                bank.get_x_coordinate("thirdparty") % 10,
                bank.get_y_coordinate("thirdparty") % 10,
                bank.get_x_coordinate("city") / 10,
                bank.get_x_coordinate("day") / 10,
                bank.get_x_coordinate("day") % 10,
                bank.get_x_coordinate("month") / 10,
                bank.get_x_coordinate("month") % 10,
                bank.get_x_coordinate("year") % 10]
        data = "".join([str(d) for d in data])
        self.write("%c%c987? %s%c" % (DP20C.CMD_PREFIX,
                                      DP20C.CMD_SETUP_COORDINATES,
                                      data, DP20C.CMD_SUFFIX))

    def _setup_cheque(self, bank, value, thirdparty, city,
                      date=None):
        if date is None:
            date = datetime.datetime.now()
        self._setup_positions(bank)
        value = "%.02f" % value
        if "." in value:
            value.replace(".", ",")
        date = date.strftime("%02d/%02m/%Y")
        bank_code = "987"
        for idx, data in enumerate((thirdparty, city, bank_code, value, date)):
            s = "%c%s" % (0xA0 + idx, data)
            self.write(DP20C.CMD_PREFIX + s + DP20C.CMD_SUFFIX)

    def _print_cheque(self):
        for data in (0xB1, 0xB0):
            self.write(DP20C.CMD_PREFIX + chr(data))

    #
    # IChequePrinter implementation
    #

    def print_cheque(self, *args, **kwargs):
        self._setup_cheque(*args, **kwargs)
        self._print_cheque()

    def get_capabilities(self):
        # XXX: The Bematech DP20C manual doesn't specify what are the
        # parameter max values, so...
        return {'cheque_thirdparty': Capability(),
                'cheque_value': Capability(),
                'cheque_city': Capability()}



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Johan Dahlin    <jdahlin@async.com.br>
##

# This module implements the ABICOMP codec for python

TABLE = {
    u'À': '\xa1',
    u'Á': '\xa2',
    u'Â': '\xa3',
    u'Ã': '\xa4',
    u'Ä': '\xa5',
    u'Ç': '\xa6',
    u'È': '\xa7',
    u'É': '\xa8',
    u'Ê': '\xa9',
    u'Ë': '\xaa',
    u'Ì': '\xab',
    u'Í': '\xac',
    u'Î': '\xad',
    u'Ï': '\xae',
    u'Ñ': '\xaf',

    u'Ò': '\xb0',
    u'Ó': '\xb1',
    u'Ô': '\xb2',
    u'Õ': '\xb3',
    u'Ö': '\xb4',
    u'Œ': '\xb5',
    u'Ù': '\xb6',
    u'Ú': '\xb7',
    u'Û': '\xb8',
    u'Ü': '\xb9',
    u'Ÿ': '\xba',
    u'˝': '\xbb',
    u'£': '\xbc',
    u'ʻ': '\xbd',
    u'°': '\xbe',

    u'¡': '\xc0',
    u'à': '\xc1',
    u'á': '\xc2',
    u'â': '\xc3',
    u'ã': '\xc4',
    u'ä': '\xc5',
    u'ç': '\xc6',
    u'è': '\xc7',
    u'é': '\xc8',
    u'ê': '\xc9',
    u'ë': '\xca',
    u'ì': '\xcb',
    u'í': '\xcc',
    u'î': '\xcd',
    u'ï': '\xce',
    u'ñ': '\xcf',

    u'ò': '\xd0',
    u'ó': '\xd1',
    u'ô': '\xd2',
    u'õ': '\xd3',
    u'ö': '\xd4',
    u'œ': '\xd5',
    u'ù': '\xd6',
    u'ú': '\xd7',
    u'û': '\xd8',
    u'ü': '\xd9',
    u'ÿ': '\xda',
    u'ß': '\xdb',
    u'ª': '\xdc',
    u'º': '\xdd',
    u'¿': '\xde',
    u'±': '\xdf',
}
RTABLE = dict([(v, k) for k, v in TABLE.items()])


def encode(input):
    """
    Convert unicode to string.
    @param input: text to encode
    @type input: unicode
    @returns: encoded text
    @rtype: str
    """
    return [TABLE.get(c) or str(c) for c in input]


def decode(input):
    """
    Convert string in unicode.
    @param input: text to decode
    @type input: str
    @returns: decoded text
    @rtype: unicode
    """
    return [RTABLE.get(c) or unicode(c) for c in input]


def register_codec():
    import codecs

    class Codec(codecs.Codec):
        def encode(self, input, errors='strict'):
            if not input:
                return "", 0
            output = encode(input)
            return "".join(output), len(output)

        def decode(self, input, errors='strict'):
            if not input:
                return u"", 0
            output = decode(input)
            return u"".join(output), len(output)

    class StreamWriter(Codec, codecs.StreamWriter):
        pass

    class StreamReader(Codec, codecs.StreamReader):
        pass

    def getregentry(encoding):
        if encoding != 'abicomp':
            return None
        return (Codec().encode,
                Codec().decode,
                StreamReader,
                StreamWriter)

    codecs.register(getregentry)


def test():
    register_codec()
    all = u''.join(TABLE.keys())
    assert all == unicode(all.encode('abicomp'), 'abicomp')

    mixed = u'não dîz'
    assert mixed == unicode(mixed.encode('abicomp'), 'abicomp')

if __name__ == '__main__':
    test()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano  <henrique@async.com.br>
##
"""
Base class implementation for all the scales drivers.
"""

from stoqdrivers.scales.base import BaseScale

#
# Scale interface
#


class Scale(BaseScale):
    def read_data(self):
        return self._driver.read_data()


def test():
    scale = Scale()

    print "Waiting for scale reply... "
    data = scale.read_data()
    print "...ok"
    print "Weight: %.02f" % data.weight
    print "Price per Kg: %.02f" % data.price_per_kg
    print "Total price: %.02f" % data.total_price

if __name__ == "__main__":
    test()



# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005-2011 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Henrique Romano <henrique@async.com.br>
##              Johan Dahlin <henrique@async.com.br>
##

__version__ = (0, 9, 24)

from kiwi.environ import Library

__all__ = ["library"]

library = Library("stoqdrivers", root="..")
library.enable_translation()



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2007 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Johan Dahlin                <jdahlin@async.com.br>
##
"""
StoqDrivers enums
"""

from kiwi.python import enum


class PaymentMethodType(enum):
    """
    Enum for Payment Methods
    """
    (MONEY,
     CHECK,
     BILL,
     CREDIT_CARD,
     DEBIT_CARD,
     FINANCIAL,
     GIFT_CERTIFICATE,
     CUSTOM,
     MULTIPLE) = range(9)


class UnitType(enum):
    """
    Enum for units
    """
    (WEIGHT,
     METERS,
     LITERS,
     EMPTY,
     CUSTOM) = range(20, 25)


class TaxType(enum):
    """
    Enum for taxes
    """
    (ICMS,
     SUBSTITUTION,
     EXEMPTION,
     NONE,
     SERVICE,
     CUSTOM) = range(40, 46)


class DeviceType(enum):
    """
    Enum for device types
    """

    (PRINTER,
     SCALE,
     BARCODE_READER) = range(3)



# -*- Mode: Python; coding: iso-8859-1 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Stoqdrivers
## Copyright (C) 2005 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
## USA.
##
## Author(s):   Evandro Vale Miquelito      <evandro@async.com.br>
##              Henrique Romano             <henrique@async.com.br>
##
"""
Useful routines when parsing the configuration file
"""
import os
from ConfigParser import ConfigParser

from stoqdrivers.exceptions import ConfigError
from stoqdrivers.translation import stoqdrivers_gettext

_ = stoqdrivers_gettext


class StoqdriversConfig:

    domain = 'stoqdrivers'

    def __init__(self, filename=None):
        """ filename is the name of the configuration file we're reading """

        self.filename = filename or (self.domain + '.conf')
        self.config = ConfigParser()
        self._load_config()

    def get_homepath(self):
        return os.path.join(os.getenv('HOME'), '.' + self.domain)

    def _open_config(self, path):
        filename = os.path.join(path, self.filename)
        if not os.path.exists(filename):
            return False
        self.config.read(filename)
        return True

    def _load_config(self):
        # Try to load configuration  from:
        # 1) $HOME/.$domain/$filename
        # 2) $PREFIX/etc/$domain/$filename
        # 3) /etc/$filename

        # This is a bit hackish:
        # $prefix / lib / $DOMAIN / lib / config.py
        #    -4      -3     -2      -1       0
        filename = os.path.abspath(__file__)
        stripped = filename.split(os.sep)[:-4]
        self.prefix = os.path.join(os.sep, *stripped)

        homepath = self.get_homepath()
        etcpath = os.path.join(self.prefix, 'etc', self.domain)
        globetcpath = os.path.join(os.sep, 'etc', self.domain)
        if not (self._open_config(homepath) or self._open_config(etcpath) or
                self._open_config(globetcpath)):
            raise ConfigError(_("Config file not found in: `%s', `%s' and "
                                "`%s'") % (homepath, etcpath, globetcpath))

    def has_section(self, section):
        return self.config.has_section(section)

    def has_option(self, name, section='General'):
        return self.config.has_option(section, name)

    def get_option(self, name, section='General'):
        if not self.config.has_section(section):
            raise ConfigError(_("Invalid section: %s") % section)
        elif not self.config.has_option(section, name):
            raise ConfigError(_("%s does not have option: %s")
                              % (self.filename, name))
        return self.config.get(section, name)

    def set_option(self, name, section='General'):
        if not self.config.has_section(section):
            raise ConfigError(_("Invalid section: %s") % section)
        self.config.set(section, name)
